/*--------------------------------------------------------------------------
Functions to manipulate directory information
--------------------------------------------------------------------------*/
#include <list>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <dirent.h>
#include "nfs.h"
#include "nfsd.h"

using namespace std;

extern "C" int buildDirEntries(const char *fullPath,
			       const nfscookie *cookie,
			       unsigned maxBytes);
extern "C" struct entry *getStartEntry();
extern "C" int getEOF();

static off_t getOffsetFromCookie(const nfscookie *cookie);
static void storeOffsetToCookie(nfscookie *cookie, off_t offset);
static unsigned getFileID(const char *name);
static void clearEntryPool();
static void addToPool(const struct entry *nextEntry);
static void printEntries(struct entry* e, int full);
static int eof = 1;
static list<struct entry> entryPool;

/*--------------------------------------------------------------------------
Build directory entries
--------------------------------------------------------------------------*/
int buildDirEntries(const char *fullPath,
		    const nfscookie *cookie,
		    unsigned maxBytes)
{
  int len = 0;
  DIR *dir = NULL;
  int sumBytes = 0;
  int d_name_len = 0;
  dirent *ent = NULL;
  struct entry nextEntry;
  off_t offset = (off_t)0;
  char fullFileName[MAX_NFS_PATH];

  offset = getOffsetFromCookie(cookie);
  dir = opendir(fullPath);
  if(dir == NULL){
    return(0);
  }
  if(offset != 0){
    seekdir(dir, offset);
  }
  sumBytes = 0;
  strcpy(fullFileName, fullPath);
  len = strlen(fullFileName);
  if(fullFileName[len - 1] != '/'){
    fullFileName[len] = '/';
    fullFileName[len + 1] = '\0';
    len++;
  }
  clearEntryPool();
  eof = 0;
  ent = readdir(dir);
  while((ent != NULL) && (sumBytes <= maxBytes)){
    strcpy(fullFileName + len, ent->d_name);
    nextEntry.nextentry = NULL;
    nextEntry.fileid = getFileID(fullFileName);
    d_name_len = strlen(ent->d_name);
    nextEntry.name = (char *)malloc(d_name_len + 1);
    strcpy(nextEntry.name, ent->d_name);
    offset = telldir(dir);
    if(offset == -1){
      return(0);
    }
    storeOffsetToCookie(&nextEntry.cookie, offset);
    sumBytes += sizeof(struct entry) + d_name_len;
    if(sumBytes <= maxBytes){
      addToPool(&nextEntry);
      ent = readdir(dir);
    }
  }
  eof = (ent == NULL);
  closedir(dir);
  return(1);
}

/*--------------------------------------------------------------------------
Convert a cookie into an off_t
--------------------------------------------------------------------------*/
off_t getOffsetFromCookie(const nfscookie *cookie)
{
  off_t offset = (off_t)0;

  for(int i = 0; i < COOKIESIZE; i++){
    offset = offset << 8;
    offset = offset | (cookie->data[COOKIESIZE - i - 1] & 255);
  }
  return(offset);
}

/*--------------------------------------------------------------------------
Store an off_t into a cookie
--------------------------------------------------------------------------*/
void storeOffsetToCookie(nfscookie *cookie, off_t offset)
{
  for(int i = 0; i < COOKIESIZE; i++){
    cookie->data[i] = (char)(offset & ((char)-1));
    offset = offset >> 8;
  }
}

/*--------------------------------------------------------------------------
Use the inode number of a file as its file id
--------------------------------------------------------------------------*/
unsigned getFileID(const char* name)
{
  int retVal = 0;
  unsigned id = 0;
  struct stat fileInfo;

  retVal = stat(name, &fileInfo);
  if(retVal == 0){
    id = (unsigned)fileInfo.st_ino;
  }else{
    id = (unsigned)~0;
  }
  return(id);
}

/*--------------------------------------------------------------------------
Clear the entry pool
--------------------------------------------------------------------------*/
void clearEntryPool()
{
  for(list<struct entry>::iterator i = entryPool.begin();
      i != entryPool.end();
      i++){
    free((*i).name);
  }
  entryPool.clear();
}

/*--------------------------------------------------------------------------
Clear the entry pool
--------------------------------------------------------------------------*/
void addToPool(const struct entry *nextEntry)
{
  list<struct entry>::reverse_iterator index;
  list<struct entry>::reverse_iterator previous;

  entryPool.push_back(*nextEntry);
  index = entryPool.rbegin();
  assert(!(index == entryPool.rend()));
  previous = index;
  previous++;
  if(previous != entryPool.rend()){
    previous->nextentry = &(*index);
  }
}

/*--------------------------------------------------------------------------
Return a pointer to the first entry
--------------------------------------------------------------------------*/
struct entry *getStartEntry()
{
  struct entry *ptr = NULL;

  if(!entryPool.empty()){
    ptr = &(entryPool.front());
  }
  return(ptr);
}

/*--------------------------------------------------------------------------
Return eof
--------------------------------------------------------------------------*/
int getEOF()
{
  return(eof);
}

/*--------------------------------------------------------------------------
Used for testing; print the list of entries
--------------------------------------------------------------------------*/
void printEntries(struct entry *e, int full)
{
  while(e != NULL){
    if(full){
      printf("--->\n");
      printf("fileID = %lu\n", e->fileid);
      printf("name = %s\n", e->name);
      printf("cookie.data = [%d, %d, %d, %d]\n",
	     (e->cookie.data[3] & 255),
	     (e->cookie.data[2] & 255),
	     (e->cookie.data[1] & 255),
	     (e->cookie.data[0] & 255));
      printf("(offset) = %lu\n", getOffsetFromCookie(&e->cookie));
    }else{
      printf("%s\n", e->name);
    }
    e = e->nextentry;
  }
}

/*--------------------------------------------------------------------------
Used for testing; remove when compiling all
--------------------------------------------------------------------------*/
/*
int main()
{
  int ret = 0;
  nfscookie cookie;

  cookie.data[0] = (char)0;
  cookie.data[1] = (char)0;
  cookie.data[2] = (char)0;
  cookie.data[3] = (char)0;
  do{
    ret = buildDirEntries("/var/local/", &cookie, 10000);
    assert(ret);
    if(!entryPool.empty()){
      printEntries(&(entryPool.front()), false);
    }
    if(!eof){
      cookie.data[0] = (entryPool.back()).cookie.data[0];
      cookie.data[1] = (entryPool.back()).cookie.data[1];
      cookie.data[2] = (entryPool.back()).cookie.data[2];
      cookie.data[3] = (entryPool.back()).cookie.data[3];
    }
  }while(!eof);
  return(0);
}
*/
