/*--------------------------------------------------------------------------
Functions to manipulate fhandle
--------------------------------------------------------------------------*/
#include <assert.h>
#include <stdio.h>
#include <string.h>
#include <vector>
#include "nfs.h"
#include "nfsd.h"

/* If component smaller than this, don't replace with number */
#define MIN_NUM_CONV_LEN 3

/* NOTE: Testing */
/*
#undef FHSIZE
#define FHSIZE 20
*/

using namespace std;

static vector<char *> v;
static FILE *fhTable;

#ifdef __cplusplus
extern "C"{
#endif

  void initHandleTable(const char* fhTableName);
  int fhandleToClientName(const fhandle *handle, char* name);
  int clientNameTofhandle(const char* nm, fhandle *handle);

#ifdef __cplusplus
}
#endif

static char *indexOf(char* str, char ch);
static void getNameFromNum(int num, char *newName);
static void shiftRight(char *str, int shiftAmt);
static void shiftLeft(char *str, int shiftAmt);
static void replace(char *oldStr, char *newStr);
static void addEscapes(char *name);
static int getNumFromName(char* name);

/*--------------------------------------------------------------------------
Initialize the table
--------------------------------------------------------------------------*/
void initHandleTable(const char *fhTableName)
{
  FILE *file = NULL;
  char *entry = NULL;
  char *entryCopy = NULL;
  char buffer[NFS_MAXPATHLEN + 1];

  file = fopen(fhTableName, "r");
  if(file != NULL){
    entry = fgets(buffer, NFS_MAXPATHLEN, file);
    while(entry != NULL){
      /* Add the entry */
      if(entry[strlen(entry) - 1] == '\n'){
	entry[strlen(entry) - 1] = '\0';
      }
      entryCopy = new char[strlen(entry) + 1];
      strcpy(entryCopy, entry);
      v.push_back(entryCopy);
      entry = fgets(buffer, NFS_MAXPATHLEN, file);
    }
    fclose(file);
  }

  fhTable = fopen(fhTableName, "a");
  assert(fhTable != NULL);
}

/*--------------------------------------------------------------------------
Convert from a fhandle to a real file name
--------------------------------------------------------------------------*/
int fhandleToClientName(const fhandle *handle, char* name)
{
  int num = 0;
  int numStrLen = 0;
  char newName[256];
  int newNameLen = 0;
  char tempCh = '\0';
  char *endIndex = NULL;
  char *startIndex = NULL;
  
  strcpy(name, handle->data);
  startIndex = name;
  while((*startIndex) != '\0'){
    endIndex = indexOf(startIndex + 1, '/');
    if(*startIndex == '@'){
        if(*(startIndex + 1) != '@'){
	  tempCh = *endIndex;
	  *endIndex = '\0';
	  num = atoi(startIndex + 1);
	  numStrLen = strlen(startIndex + 1);
	  *endIndex = tempCh;
	  getNameFromNum(num, newName);
	  newNameLen = strlen(newName);
	  if(newNameLen > (numStrLen + 1)){   // +1 for '@'
	    shiftRight(endIndex, newNameLen - numStrLen - 1);
	  }else{
	    shiftLeft(endIndex, numStrLen - newNameLen + 1);
	  }
	  assert(strlen(name) < MAX_NFS_PATH);
	  replace(startIndex, newName);
	  endIndex += (newNameLen - numStrLen - 1);
	}else{
	  shiftLeft(startIndex + 1, 1);
	  assert(strlen(name) < MAX_NFS_PATH);
	  endIndex -= 1;
	}
	assert((*endIndex == '/') || (*endIndex == '\0'));
      }
    startIndex = endIndex + 1;
  }
  return(1);
}

/*--------------------------------------------------------------------------
Return a pointer to the location where ch lies in str
--------------------------------------------------------------------------*/
static char *indexOf(char* str, char ch)
{
  char *endIndex = NULL;

  for(endIndex = str; *endIndex != '\0'; endIndex++){
    if(*endIndex == ch){
      break;
    }
  }
  return(endIndex);
}

/*--------------------------------------------------------------------------
Get the name for which num is a shortcut
--------------------------------------------------------------------------*/
void getNameFromNum(int num, char *newName)
{
  assert(num < v.size());
  strcpy(newName, v[num]);
}

/*--------------------------------------------------------------------------
Move the text in the string right by shiftAmt
--------------------------------------------------------------------------*/
void shiftRight(char *str, int shiftAmt)
{
  char *nextChar = NULL;

  // Want to move NULL char also
  for(nextChar = str + strlen(str); nextChar >= str; nextChar--){
    *(nextChar + shiftAmt) = *nextChar;
  }
}

/*--------------------------------------------------------------------------
Move the text in the string left by shiftAmt
--------------------------------------------------------------------------*/
void shiftLeft(char *str, int shiftAmt)
{
  int len = 0;
  char *nextChar = NULL;

  len = strlen(str);
  // Want to move NULL char also
  for(nextChar = str; nextChar <= str + len; nextChar++){
    *(nextChar - shiftAmt) = *nextChar;
  }
}

/*--------------------------------------------------------------------------
Replace text in oldStr by text in newStr
--------------------------------------------------------------------------*/
void replace(char *oldStr, char *newStr)
{
  int len = 0;
  char ch = '\0';

  len = strlen(newStr);
  ch = *(oldStr + len);
  strcpy(oldStr, newStr);
  *(oldStr + len) = ch;
}

/*--------------------------------------------------------------------------
Convert from a real file name to a fhandle
--------------------------------------------------------------------------*/
int clientNameTofhandle(const char* nm, fhandle *handle)
{
  int num = 0;
  int success = 0;
  char numStr[10];
  int numStrLen = 0;
  int oldStrLen = 0;
  char tempCh = '\0';
  char *endIndex = NULL;
  char *startIndex = NULL;
  char name[MAX_NFS_PATH];

  success = 0;
  strcpy(name, nm);
  addEscapes(name);
  if(strlen(name) >= FHSIZE){
    startIndex = name;
    while(*startIndex != '\0'){
      endIndex = indexOf(startIndex + 1, '/');
      if((*startIndex != '@') || (*(startIndex + 1) != '@')){
	// Found a non-shortcut
	tempCh = *endIndex;
	*endIndex = '\0';
	oldStrLen = strlen(startIndex);
	if(oldStrLen > MIN_NUM_CONV_LEN){
	  // Long enough to compress
	  num = getNumFromName(startIndex);
	  *endIndex = tempCh;
	  sprintf(numStr, "@%d", num);
	  numStrLen = strlen(numStr);
	  if(oldStrLen > numStrLen){
	    shiftLeft(endIndex, oldStrLen - numStrLen);
	  }else{
	    shiftRight(endIndex, numStrLen - oldStrLen);
	  }
	  replace(startIndex, numStr);
	  endIndex -= (oldStrLen - numStrLen);
	}else{
	  *endIndex = tempCh;
	}
      }
      startIndex = endIndex + 1;
      if(strlen(name) < FHSIZE){
	success = 1;
	// End the loop
	startIndex += strlen(startIndex);
	strcpy(handle->data, name);
      }
    }
  }else{
    // File name short enough to not convert to handle
    strcpy(handle->data, name);
    success = 1;
  }

  return(success);
}

/*--------------------------------------------------------------------------
Replace espace character with escape sequence (twice escape character)
--------------------------------------------------------------------------*/
void addEscapes(char *name)
{
  while(*name != '\0'){
    if(*name == '@'){
      shiftRight(name, 1);
      *name = '@';
    }
    for(; ((*name != '/') && (*name != '\0')); name++);
    if(*name == '/'){
      name++;
    }
  }
}

/*--------------------------------------------------------------------------
Get (generate if not found) a shortcut num from the name
--------------------------------------------------------------------------*/
int getNumFromName(char* name)
{
  int num = -1;
  char *nameCopy = NULL;

  for(int i = 0; i < v.size(); i++){
    if(strcmp(name, v[i]) == 0){
      num = i;
      break;
    }
  }
  if(num < 0){
    nameCopy = new char[strlen(name) + 1];
    strcpy(nameCopy, name);
    v.push_back(nameCopy);
    fprintf(fhTable, "%s\n", nameCopy);
    fflush(fhTable);
    num = v.size() - 1;
  }
  return(num);
}

/*--------------------------------------------------------------------------
Testing only; must remove when compiling everything
--------------------------------------------------------------------------*/
/*
int main()
{
  char name[1024];
  fhandle f;

  initHandleTable("testfile");
  //strcpy(name, "something/vat/vat/something/@");
  strcpy(name, "something/vat/vat/something/@");
  printf("Original = %s\n", name);
  int b = clientNameTofhandle(name, &f);
  assert(b);
  printf("Got back: %s\n", f.data);

  //strcpy(f.data, "something/@0/@0/something/@@");
  fhandleToClientName(&f, name);
  printf("name = %s\n", name);

  return(0);
}
*/
