/*
 * Manipulate attributes
 */

#include "nfs.h"
#include "nfsd.h"
#include "nfs_interface.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/unistd.h>
#include <sys/statfs.h>
#include <errno.h>
#include <assert.h>
#include <utime.h>

#ifdef VERBOSE
#define debug_perror(a) perror(a)
#else
#define debug_perror(a)
#endif

/* NOTE: Temporary */
#include <time.h>

extern int errno;
#ifdef __cplusplus
extern "C"{
#endif

  int fhandleToClientName(const fhandle *handle, char* name);
  int clientNameTofhandle(const char* nm, fhandle *handle);

  int buildDirEntries(const char *fullPath,
		      const nfscookie *cookie,
		      unsigned maxBytes);
  struct entry *getStartEntry();
  int getEOF();

#ifdef __cplusplus
}
#endif

#define READ_OWNER 0000400
#define WRITE_OWNER 0000200
#define ROOT_DIR "/home/root/root_dir/"
#define ROOT_DIR_NAME_LEN 20

/*--------------------------------------------------------------------------
  Return true if this is a handle with 0's only
--------------------------------------------------------------------------*/
int zeroHandle(fhandle *argp)
{
  int allZeroes = 1;

  for(int i = 0; i < FHSIZE; i++){
    if(argp->data[i] != (char)0){
      allZeroes = 0;
      break;
    }
  }

  return(allZeroes);
}

/*--------------------------------------------------------------------------
  Translate to NFS error messages
--------------------------------------------------------------------------*/
nfsstat translateError(int code)
{
  nfsstat retVal = NFS_OK;

  switch(code){
  case 0:
    retVal = NFS_OK;
    break;
  case ENOENT:
    retVal = NFSERR_NOENT;
    break;
  case ENOTDIR:
    retVal = NFSERR_NOTDIR;
    break;
  case ENAMETOOLONG:
    retVal = NFSERR_NAMETOOLONG;
    break;
  case EPERM:
    retVal = NFSERR_PERM;
    break;
  case EROFS:
    retVal = NFSERR_ROFS;
    break;
  case EFAULT:
    retVal = NFSERR_ACCES;  /* NOTE: Approximation */
    break;
  case EACCES:
    retVal = NFSERR_ACCES;
    break;
  case EIO:
    retVal = NFSERR_IO;
    break;
  case ENOTEMPTY:
    retVal = NFSERR_NOTEMPTY;
    break;
  case ELOOP:
    retVal = NFSERR_NOENT;  /* NOTE: Approximation */
    break;
  default:
    retVal = NFSERR_IO;
    break;
  }
  return(retVal);
}

/*--------------------------------------------------------------------------
  Fill in attribute fields
--------------------------------------------------------------------------*/
int fillAttributes(const char *fullFileName, fattr *attributes)
{
  int retVal = 0;
  int success = 0;
  struct stat fileInfo;

  errno = 0;
  retVal = lstat(fullFileName, &fileInfo);
  if(retVal != 0){
    return(errno);
  }

  if(S_ISDIR(fileInfo.st_mode)){
    attributes->type = NFDIR;
  }else if(S_ISREG(fileInfo.st_mode)){
    attributes->type = NFREG;
  }else if(S_ISLNK(fileInfo.st_mode)){
    attributes->type = NFLNK;
  }else{
    return(ENOTBLK);
  }

  attributes->mode = fileInfo.st_mode;
  attributes->nlink = fileInfo.st_nlink;
  attributes->uid = (u_int)fileInfo.st_uid;
  attributes->gid = (u_int)fileInfo.st_gid;
  attributes->size = (u_int)fileInfo.st_size;
  attributes->blocksize = (u_int)fileInfo.st_blksize;
  attributes->rdev = (u_int)fileInfo.st_rdev;
  attributes->blocks = (u_int)fileInfo.st_blocks;
  /* attributes->fsid = (u_int)fsInfo.f_fsid; */
  attributes->fsid = (u_int)FSID;
  attributes->fileid = (u_int)fileInfo.st_ino;
  attributes->atime.seconds = (u_int)fileInfo.st_atime;
  attributes->mtime.seconds = (u_int)fileInfo.st_mtime;
  attributes->ctime.seconds = (u_int)fileInfo.st_ctime;
  attributes->atime.useconds = (u_int)0;
  attributes->mtime.useconds = (u_int)0;
  attributes->ctime.useconds = (u_int)0;

  return(0);
}

/*--------------------------------------------------------------------------
  Change attributes
--------------------------------------------------------------------------*/
int changeAttributes(const char* fullPath, sattr *attributes)
{
  int retVal = 0;
  struct stat fileStats;
  struct utimbuf timeBuffer;

  if(attributes->mode != (unsigned)~0){
    retVal = chmod(fullPath, attributes->mode);
    if(retVal != 0){
      debug_perror("changeAttributes: chmod");
      return(errno);
    }
  }

  retVal = chown(fullPath, (uid_t)attributes->uid, (gid_t)attributes->gid);
  if(retVal != 0){
    debug_perror("changeAttributes: chown");
    return(errno);
  }

  if(attributes->size != (unsigned)~0){
    retVal = truncate(fullPath, (off_t)attributes->size);
    if(retVal != 0){
      debug_perror("changeAttributes: truncate");
      return(errno);
    }
  }

  retVal = lstat(fullPath, &fileStats);
  if(retVal != 0){
    debug_perror("changeAttributes: stat");
    return(errno);
  }

  timeBuffer.actime = attributes->atime.seconds;
  timeBuffer.modtime = attributes->mtime.seconds;
  if(attributes->atime.seconds == (unsigned)~0){
    timeBuffer.actime = fileStats.st_atime;
  }
  if(attributes->mtime.seconds == (unsigned)~0){
    timeBuffer.modtime = fileStats.st_mtime;
  }
  retVal = utime(fullPath, &timeBuffer);
  if(retVal != 0){
    debug_perror("changeAttributes: utime");
    return(errno);
  }
  return(0);
}

/*--------------------------------------------------------------------------
  Fill the result with correct file attributes
--------------------------------------------------------------------------*/
void getattr(attrstat *result, fhandle *argp, struct svc_req *rqstp)
{
  int retVal = 0;
  char clientName[MAX_NFS_PATH];

  result->status = NFS_OK;

  /* If we a get a 0 string file handle, treat it separately */
  strcpy(clientName, ROOT_DIR);
  if(zeroHandle(argp)){
    /* Root directory */
    retVal = 1;
  }else{
    retVal = fhandleToClientName(argp, clientName + ROOT_DIR_NAME_LEN);
    if(!retVal){
      result->status = NFSERR_STALE;
      return;
    }
  }
  retVal = fillAttributes(clientName, &result->attributes);
  if(retVal != 0){
    debug_perror("getattr");
    result->status = translateError(retVal);
    return;
  }else{
    result->status = NFS_OK;
  }
}

/*--------------------------------------------------------------------------
  Return the contents of a directory
--------------------------------------------------------------------------*/
void readdir(readdirres *result, readdirargs *argp, struct svc_req *rqstp)
{
  int retVal = 0;
  fhandle handle;
  char fullPath[MAX_NFS_PATH + 25]; // ROOT_DIR included

  strcpy(fullPath, ROOT_DIR);
  if(zeroHandle(&argp->dir)){
    retVal = 1;
  }else{
    retVal = fhandleToClientName(&argp->dir, fullPath + ROOT_DIR_NAME_LEN);
    if(!retVal){
      result->status = NFSERR_STALE;
      return;
    }
  }
  retVal = buildDirEntries(fullPath, &argp->cookie, argp->count);
  if(retVal){
    result->readdirok.entries = getStartEntry();
    result->readdirok.eof = getEOF();
    result->status = NFS_OK;
  }else{
    result->status = translateError(errno);
  }
}

/*--------------------------------------------------------------------------
  Fill the result with correct file system attributes
--------------------------------------------------------------------------*/
void statfs(statfsres *result, fhandle *argp, struct svc_req *rqstp)
{
  int retVal = 0;
  struct statfs fsinfo;

  retVal = statfs(ROOT_DIR, &fsinfo);
  if(retVal != 0){
    result->status = translateError(errno);
    return;
  }

  result->info.tsize = (u_int)fsinfo.f_bsize; /* NOTE: Arbitrary */
  result->info.bsize = (u_int)fsinfo.f_bsize;
  result->info.blocks = (u_int)fsinfo.f_blocks;
  result->info.bfree = (u_int)fsinfo.f_bfree;
  result->info.bavail = (u_int)fsinfo.f_bavail;
  result->status = NFS_OK;
}

/*--------------------------------------------------------------------------
  Look up a file in a directory
--------------------------------------------------------------------------*/
void lookup(diropres *result, diropargs *argp, struct svc_req *rqstp)
{
  int len = 0;
  int retVal = 0;
  char fullPath[MAX_NFS_PATH + 25]; // ROOT_DIR included

  strcpy(fullPath, ROOT_DIR);
  if(zeroHandle(&argp->dir)){
    retVal = 1;
  }else{
    retVal = fhandleToClientName(&argp->dir, fullPath + ROOT_DIR_NAME_LEN);
    if(!retVal){
      result->status = NFSERR_STALE;
      return;
    }
  }
  len = strlen(fullPath);
  if(fullPath[len - 1] != '/'){
    fullPath[len] = '/';
    fullPath[len + 1] = '\0';
  }
  strcat(fullPath, argp->name);

  retVal = fillAttributes(fullPath, &result->diropok.attributes);
  if(retVal != 0){
    debug_perror("lookup");
    result->status = translateError(retVal);
    return;
  }
  retVal = clientNameTofhandle(fullPath + ROOT_DIR_NAME_LEN,
			       &result->diropok.file);
  if(!retVal){
    result->status = NFSERR_NAMETOOLONG;
    return;
  }
  result->status = NFS_OK;
}

/*--------------------------------------------------------------------------
  Read from a file
--------------------------------------------------------------------------*/
void read(readres *result, readargs *argp, struct svc_req *rqstp)
{
  int retVal = 0;
  FILE *file = NULL;
  char fullPath[MAX_NFS_PATH + 25]; // ROOT_DIR included
  static char *buffer = NULL; // Lifetime beyond function

  strcpy(fullPath, ROOT_DIR);
  if(zeroHandle(&argp->file)){
    retVal = 1;
  }else{
    retVal = fhandleToClientName(&argp->file, fullPath + ROOT_DIR_NAME_LEN);
    if(!retVal){
      result->status = NFSERR_STALE;
      return;
    }
  }
  file = fopen(fullPath, "r");
  if(file == NULL){
    result->status = translateError(errno);
    return;
  }

  retVal = fseek(file, argp->offset, SEEK_SET);
  if(retVal != 0){
    result->status = translateError(errno);
    fclose(file);
    return;
  }

  if(buffer != NULL){
    free(buffer);
    buffer = NULL;
  }
  buffer = (char*)malloc(argp->count);
  result->reply.data.data_len = fread(buffer, 1, argp->count, file);
  if(ferror(file)){
    result->status = translateError(errno);
    fclose(file);
    return;
  }
  result->reply.data.data_val = buffer;
  fclose(file);

  retVal = fillAttributes(fullPath, &result->reply.attributes);
  if(retVal != 0){
    debug_perror("read");
    result->status = translateError(retVal);
  }

  result->status = NFS_OK;
}

/*--------------------------------------------------------------------------
  Set attributes
--------------------------------------------------------------------------*/
void setattr(attrstat *result, sattrargs *argp, struct svc_req *rqstp)
{
  int retVal = 0;
  char fullPath[MAX_NFS_PATH + 25]; /* ROOT_DIR included */

  strcpy(fullPath, ROOT_DIR);
  if(zeroHandle(&argp->file)){
    retVal = 1;
  }else{
    retVal = fhandleToClientName(&argp->file, fullPath + ROOT_DIR_NAME_LEN);
    if(!retVal){
      result->status = NFSERR_STALE;
      return;
    }
  }
  retVal = changeAttributes(fullPath, &argp->attributes);
  if(retVal != 0){
    result->status = translateError(retVal);
    return;
  }

  retVal = fillAttributes(fullPath, &result->attributes);
  if(retVal != 0){
    debug_perror("setattr");
    result->status = translateError(retVal);
  }else{
    result->status = NFS_OK;
  }
}

/*--------------------------------------------------------------------------
  Create a new file
--------------------------------------------------------------------------*/
void create(diropres *result, createargs *argp, struct svc_req *rqstp)
{
  int len = 0;
  int retVal = 0;
  FILE *file = NULL;
  char fullPath[MAX_NFS_PATH + 25]; /* ROOT_DIR included */

  strcpy(fullPath, ROOT_DIR);
  if(zeroHandle(&argp->where.dir)){
    retVal = 1;
  }else{
    retVal = fhandleToClientName(&argp->where.dir,
				 fullPath + ROOT_DIR_NAME_LEN);
    if(!retVal){
      result->status = NFSERR_STALE;
      return;
    }
  }
  len = strlen(fullPath);
  if(fullPath[len - 1] != '/'){
    fullPath[len] = '/';
    fullPath[len + 1] = '\0';
  }
  strcat(fullPath, argp->where.name);

  file = fopen(fullPath, "w");
  if(file == NULL){
    result->status = translateError(errno);
    return;
  }
  fclose(file);

  retVal = changeAttributes(fullPath, &argp->attributes);
  if(retVal != 0){
    result->status = translateError(retVal);
    return;
  }

  retVal = clientNameTofhandle(fullPath + ROOT_DIR_NAME_LEN,
			       &result->diropok.file);
  if(!retVal){
    result->status = NFSERR_NAMETOOLONG;
    return;
  }

  retVal = fillAttributes(fullPath, &result->diropok.attributes);
  if(retVal != 0){
    debug_perror("create");
    result->status = translateError(retVal);
  }else{
    result->status = NFS_OK;
  }
}

/*--------------------------------------------------------------------------
  Create a new directory
--------------------------------------------------------------------------*/
void mkdir(diropres *result, createargs *argp, struct svc_req *rqstp)
{
  int len = 0;
  int retVal = 0;
  char fullPath[MAX_NFS_PATH + 25]; /* ROOT_DIR included */

  strcpy(fullPath, ROOT_DIR);
  if(zeroHandle(&argp->where.dir)){
    retVal = 1;
  }else{
    retVal = fhandleToClientName(&argp->where.dir,
				 fullPath + ROOT_DIR_NAME_LEN);
    if(!retVal){
      result->status = NFSERR_STALE;
      return;
    }
  }
  len = strlen(fullPath);
  if(fullPath[len - 1] != '/'){
    fullPath[len] = '/';
    fullPath[len + 1] = '\0';
  }
  strcat(fullPath, argp->where.name);

  retVal = mkdir(fullPath, (mode_t)argp->attributes.mode);
  if(retVal != 0){
    result->status = translateError(errno);
    return;
  }

  retVal = changeAttributes(fullPath, &argp->attributes);
  if(retVal != 0){
    result->status = translateError(retVal);
    return;
  }

  retVal = clientNameTofhandle(fullPath + ROOT_DIR_NAME_LEN,
			       &result->diropok.file);
  if(!retVal){
    result->status = NFSERR_NAMETOOLONG;
    return;
  }

  retVal = fillAttributes(fullPath, &result->diropok.attributes);
  if(retVal != 0){
    debug_perror("mkdir");
    result->status = translateError(retVal);
  }else{
    result->status = NFS_OK;
  }
}

/*--------------------------------------------------------------------------
  Write a file
--------------------------------------------------------------------------*/
void write(attrstat *result, writeargs *argp, struct svc_req *rqstp)
{
  int retVal = 0;
  FILE *file = NULL;
  int numWritten = 0;
  char fullPath[MAX_NFS_PATH + 25]; /* ROOT_DIR included */
 
  strcpy(fullPath, ROOT_DIR);
  if(zeroHandle(&argp->file)){
    retVal = 1;
  }else{
    retVal = fhandleToClientName(&argp->file, fullPath + ROOT_DIR_NAME_LEN);
    if(!retVal){
      result->status = NFSERR_STALE;
      return;
    }
  }

  file = fopen(fullPath, "a");
  if(file == NULL){
    result->status = translateError(errno);
    return;
  }

  retVal = fseek(file, argp->offset, SEEK_SET);
  if(retVal != 0){
    result->status = translateError(errno);
    fclose(file);
    return;
  }

  numWritten = fwrite(argp->data.data_val, 1, argp->data.data_len, file);
  if(numWritten < argp->data.data_len){
    result->status = translateError(errno);
    fclose(file);
    return;
  }

  fclose(file);
  retVal = fillAttributes(fullPath, &result->attributes);
  if(retVal != 0){
    debug_perror("write");
    result->status = translateError(retVal);
    return;
  }
  result->status = NFS_OK;
}

/*--------------------------------------------------------------------------
  Remove a file
--------------------------------------------------------------------------*/
void remove(nfsstat *result, diropargs *argp, struct svc_req *rqstp)
{
  int len = 0;
  int retVal = 0;
  char fullPath[MAX_NFS_PATH + 25]; /* ROOT_DIR included */

  strcpy(fullPath, ROOT_DIR);
  if(zeroHandle(&argp->dir)){
    retVal = 1;
  }else{
    retVal = fhandleToClientName(&argp->dir, fullPath + ROOT_DIR_NAME_LEN);
    if(!retVal){
      *result = NFSERR_STALE;
      return;
    }
  }
  len = strlen(fullPath);
  if(fullPath[len - 1] != '/'){
    fullPath[len] = '/';
    fullPath[len + 1] = '\0';
  }
  strcat(fullPath, argp->name);

  retVal = remove(fullPath);
  if(retVal != 0){
    *result = translateError(errno);
    return;
  }
  *result = NFS_OK;
}

/*--------------------------------------------------------------------------
  Remove a directory
--------------------------------------------------------------------------*/
void rmdir(nfsstat *result, diropargs *argp, struct svc_req *rqstp)
{
  int len = 0;
  int retVal = 0;
  char fullPath[MAX_NFS_PATH + 25]; /* ROOT_DIR included */

  strcpy(fullPath, ROOT_DIR);
  if(zeroHandle(&argp->dir)){
    retVal = 1;
  }else{
    retVal = fhandleToClientName(&argp->dir, fullPath + ROOT_DIR_NAME_LEN);
    if(!retVal){
      *result = NFSERR_STALE;
      return;
    }
  }
  len = strlen(fullPath);
  if(fullPath[len - 1] != '/'){
    fullPath[len] = '/';
    fullPath[len + 1] = '\0';
  }
  strcat(fullPath, argp->name);

  retVal = rmdir(fullPath);
  if(retVal != 0){
    *result = translateError(errno);
    return;
  }
  *result = NFS_OK;
}

/*--------------------------------------------------------------------------
  Rename a file
--------------------------------------------------------------------------*/
void rename(nfsstat *result, renameargs *argp, struct svc_req *rqstp)
{
  int len = 0;
  int retVal = 0;
  char fromFullPath[MAX_NFS_PATH + 25]; /* ROOT_DIR included */
  char toFullPath[MAX_NFS_PATH + 25]; /* ROOT_DIR included */

  strcpy(fromFullPath, ROOT_DIR);
  if(zeroHandle(&argp->from.dir)){
    retVal = 1;
  }else{
    retVal = fhandleToClientName(&argp->from.dir,
				 fromFullPath + ROOT_DIR_NAME_LEN);
    if(!retVal){
      *result = NFSERR_STALE;
      return;
    }
  }
  len = strlen(fromFullPath);
  if(fromFullPath[len - 1] != '/'){
    fromFullPath[len] = '/';
    fromFullPath[len + 1] = '\0';
  }
  strcat(fromFullPath, argp->from.name);

  strcpy(toFullPath, ROOT_DIR);
  if(zeroHandle(&argp->to.dir)){
    retVal = 1;
  }else{
    retVal = fhandleToClientName(&argp->to.dir,
				 toFullPath + ROOT_DIR_NAME_LEN);
    if(!retVal){
      *result = NFSERR_STALE;
      return;
    }
  }
  len = strlen(toFullPath);
  if(toFullPath[len - 1] != '/'){
    toFullPath[len] = '/';
    toFullPath[len + 1] = '\0';
  }
  strcat(toFullPath, argp->to.name);

  retVal = rename(fromFullPath, toFullPath);
  if(retVal != 0){
    *result = translateError(errno);
    return;
  }
  *result = NFS_OK;
}

/*--------------------------------------------------------------------------
  Read a symbolic link
--------------------------------------------------------------------------*/
void readlink(readlinkres *result, fhandle *argp, struct svc_req *rqstp)
{
  int retVal = 0;
  int numChars = 0;
  char fakePath[MAX_NFS_PATH + 25]; // ROOT_DIR included
  static char clientPath[MAX_NFS_PATH + 25]; // ROOT_DIR included

  strcpy(fakePath, ROOT_DIR);
  if(zeroHandle(argp)){
    retVal = 1;
  }else{
    retVal = fhandleToClientName(argp, fakePath + ROOT_DIR_NAME_LEN);
    if(!retVal){
      result->status = NFSERR_STALE;
      return;
    }
  }

  numChars = readlink(fakePath, clientPath, sizeof(clientPath));
  if(numChars < 0){
    debug_perror("readlink");
    result->status = translateError(errno);
    return;
  }
  clientPath[numChars] = '\0';
  result->data = clientPath;
  result->status = NFS_OK;
}

/*--------------------------------------------------------------------------
  Make a hard link
--------------------------------------------------------------------------*/
void link(nfsstat *result, linkargs *argp, struct svc_req *rqstp)
{
  int len = 0;
  int retVal = 0;
  char toFullPath[MAX_NFS_PATH + 25]; /* ROOT_DIR included */
  char fromFullPath[MAX_NFS_PATH + 25]; /* ROOT_DIR included */

  strcpy(fromFullPath, ROOT_DIR);
  if(zeroHandle(&argp->from)){
    retVal = 1;
  }else{
    retVal = fhandleToClientName(&argp->from,
				 fromFullPath + ROOT_DIR_NAME_LEN);
    if(!retVal){
      *result = NFSERR_STALE;
      return;
    }
  }

  strcpy(toFullPath, ROOT_DIR);
  if(zeroHandle(&argp->to.dir)){
    retVal = 1;
  }else{
    retVal = fhandleToClientName(&argp->to.dir,
				 toFullPath + ROOT_DIR_NAME_LEN);
    if(!retVal){
      *result = NFSERR_STALE;
      return;
    }
  }
  len = strlen(toFullPath);
  if(toFullPath[len - 1] != '/'){
    toFullPath[len] = '/';
    toFullPath[len + 1] = '\0';
  }
  strcat(toFullPath, argp->to.name);

  retVal = link(fromFullPath, toFullPath);
  if(retVal != 0){
    debug_perror("link");
    *result = translateError(errno);
    return;
  }
  *result = NFS_OK;
}

/*--------------------------------------------------------------------------
  Make a soft link
--------------------------------------------------------------------------*/
void symlink(nfsstat *result, symlinkargs *argp, struct svc_req *rqstp)
{
  int len = 0;
  int retVal = 0;
  char newFullPath[MAX_NFS_PATH + 25]; /* ROOT_DIR included */

  strcpy(newFullPath, ROOT_DIR);
  if(zeroHandle(&argp->from.dir)){
    retVal = 1;
  }else{
    retVal = fhandleToClientName(&argp->from.dir,
				 newFullPath + ROOT_DIR_NAME_LEN);
    if(!retVal){
      *result = NFSERR_STALE;
      return;
    }
  }
  len = strlen(newFullPath);
  if(newFullPath[len - 1] != '/'){
    newFullPath[len] = '/';
    newFullPath[len + 1] = '\0';
  }
  strcat(newFullPath, argp->from.name);

  retVal = symlink(argp->to, newFullPath);
  if(retVal != 0){
    debug_perror("symlink");
    *result = translateError(errno);
    return;
  }
  *result = NFS_OK;
}
