import java.io.IOException;
import java.io.OutputStream;
import java.net.Socket;
import java.util.Enumeration;

//---------------------------------------------------------------------------
// LMHandleThread: Thread to handle requests
//---------------------------------------------------------------------------
public class LMHandleThread implements Runnable{

    //-----------------------------------------------------------------------
    // Constructor
    //-----------------------------------------------------------------------
    public LMHandleThread(SyncQueue requestQueue)
    {
	this.lockNum = 0;
	this.requestQueue = requestQueue;
	this.lockStateTable = new LMLockStateTable();
	(new Thread(this)).start();
    }

    //-----------------------------------------------------------------------
    // Main thread method
    //-----------------------------------------------------------------------
    public void run()
    {
	LMRequestPacket request = null;
	LMLockState lockState = null;

	try{
	    while(true){
		request = (LMRequestPacket)this.requestQueue.peekBlocked();
		this.requestQueue.dequeue();
		lockState = this.lockStateTable.get(request.getObjectID());

		switch(lockState.getStateName()){
		case LMLockState.IDLE:
		    this.handleIdleState(request, lockState);
		    break;
		case LMLockState.EXCLUSIVE:
		    this.handleExclusiveState(request, lockState);
		    break;
		case LMLockState.SHARED:
		    this.handleSharedState(request, lockState);
		    break;
		default:
		    assert(false);
		    break;
		}
		request = null;
		lockState = null;
	    }
	}catch(IOException e){
	    System.err.println("" + e);
	    e.printStackTrace();
	    assert(false);
	}
    }

    //-----------------------------------------------------------------------
    // Called when the lock manager FSM is in an idle state
    //-----------------------------------------------------------------------
    private void handleIdleState(LMRequestPacket request,
				 LMLockState lockState) throws IOException
    {
	int tokenID = 0;
	String objectID = null;

	//System.out.println("IDLE state");
	switch(request.getMessageType()){
	case AppMessage.LMOP_ACQUIRE_EXCLUSIVE:
	    tokenID = this.genExclusiveLockToken();
	    lockState.enqueueExclusive(request.getReplySocket(),
				       tokenID,
				       request.getSeqNum());
	    this.distributeExclusiveLock(lockState);
	    lockState.setStateName(LMLockState.EXCLUSIVE);
	    break;
	case AppMessage.LMOP_ACQUIRE_SHARED:
	    tokenID = this.genSharedLockToken();
	    lockState.enqueueShared(request.getReplySocket(),
				    tokenID,
				    request.getSeqNum());
	    this.distributeSharedLocks(lockState);
	    lockState.setStateName(LMLockState.SHARED);
	    break;
	case AppMessage.LMOP_RELEASE:    // Fall through
	case AppMessage.LMOP_UPGRADE:    // Fall through
	case AppMessage.LMOP_DOWNGRADE:  // Fall through
	default:
	    System.out.println("" + request.getMessageType());
	    assert(false);
	    break;
	}
    }

    //-----------------------------------------------------------------------
    // Called when the lock manager FSM has given out an exclusive lock
    //-----------------------------------------------------------------------
    private void handleExclusiveState(LMRequestPacket request,
				      LMLockState lockState) throws IOException
    {
	int tokenID = 0;
	String objectID = null;

	//System.out.println("EXCLUSIVE state");
	assert(lockState.hasNoOutstandingSharedLocks());
	assert(!lockState.isExclusiveQueueEmpty());
	objectID = request.getObjectID();
	switch(request.getMessageType()){
	case AppMessage.LMOP_ACQUIRE_EXCLUSIVE:
	    tokenID = this.genExclusiveLockToken();
	    lockState.enqueueExclusive(request.getReplySocket(),
				       tokenID,
				       request.getSeqNum());
	    break;
	case AppMessage.LMOP_ACQUIRE_SHARED:
	    tokenID = this.genSharedLockToken();
	    lockState.enqueueShared(request.getReplySocket(),
				    tokenID,
				    request.getSeqNum());
	    break;
	case AppMessage.LMOP_RELEASE:
	    tokenID = request.getLockToken();
	    lockState.dequeueExclusive(tokenID);
	    if(lockState.isExclusiveQueueEmpty()){
		if(lockState.isSharedQueueEmpty()){
		    // Move to the idle state by freeing lock state
		    this.lockStateTable.remove(objectID);
		}else{
		    this.distributeSharedLocks(lockState);
		    lockState.setStateName(LMLockState.SHARED);
		}
	    }else{
		this.distributeExclusiveLock(lockState);
	    }
	    break;
	case AppMessage.LMOP_DOWNGRADE:
	    tokenID = request.getLockToken();
	    lockState.dequeueExclusive(tokenID);
	    tokenID = this.genSharedLockToken();
	    lockState.enqueueShared(request.getReplySocket(),
				    tokenID,
				    request.getSeqNum());
	    if(lockState.isExclusiveQueueEmpty()){
		this.distributeSharedLocks(lockState);
		lockState.setStateName(LMLockState.SHARED);
	    }else{
		this.distributeExclusiveLock(lockState);
	    }
	    break;
	case AppMessage.LMOP_UPGRADE:    // Fall through
	default:
	    System.out.println("MessageType =  " + request.getMessageType());
	    assert(false);
	    break;
	}
    }

    //-----------------------------------------------------------------------
    // Called when the lock manager FSM has given out a shared lock
    //-----------------------------------------------------------------------
    private void handleSharedState(LMRequestPacket request,
				   LMLockState lockState) throws IOException
    {
	int tokenID = 0;
	long seqNum = 0;
	String objectID = null;
	Socket socket = null;

	//System.out.println("SHARED state");
	assert(!lockState.hasNoOutstandingSharedLocks());
	objectID = request.getObjectID();
	socket = request.getReplySocket();
	seqNum = request.getSeqNum();
	switch(request.getMessageType()){
	case AppMessage.LMOP_ACQUIRE_EXCLUSIVE:
	    tokenID = this.genExclusiveLockToken();
	    lockState.enqueueExclusive(socket, tokenID, seqNum);
	    break;
	case AppMessage.LMOP_ACQUIRE_SHARED:
	    tokenID = this.genSharedLockToken();
	    lockState.enqueueShared(socket, tokenID, seqNum);
	    if(lockState.isExclusiveQueueEmpty()){
		this.distributeSharedLocks(lockState);
	    }
	    break;
	case AppMessage.LMOP_RELEASE:
	    tokenID = request.getLockToken();
	    lockState.dequeueOutstandingSharedLock(tokenID);
	    if(lockState.hasNoOutstandingSharedLocks()){
		if(lockState.isExclusiveQueueEmpty()){
		    // Move to the idle state by freeing lock state
		    assert(lockState.isSharedQueueEmpty());
		    this.lockStateTable.remove(objectID);
		}else{
		    this.distributeExclusiveLock(lockState);
		    lockState.setStateName(LMLockState.EXCLUSIVE);
		}
	    }
	    break;
	case AppMessage.LMOP_UPGRADE:
	    tokenID = request.getLockToken();
	    lockState.dequeueOutstandingSharedLock(tokenID);
	    tokenID = this.genExclusiveLockToken();
	    lockState.enqueueExclusive(socket, tokenID, seqNum);
	    if(lockState.hasNoOutstandingSharedLocks()){
		this.distributeExclusiveLock(lockState);
		lockState.setStateName(LMLockState.EXCLUSIVE);
	    }
	    break;
	case AppMessage.LMOP_DOWNGRADE:    // Fall through
	default:
	    System.out.println("LMHandleThread: " + request.getMessageType());
	    assert(false);
	    break;
	}
    }

    //-----------------------------------------------------------------------
    // Send out a single shared lock
    //-----------------------------------------------------------------------
    private void sendSharedLock(Socket socket,
				int tokenID,
				long seqNum,
				String objectID) throws IOException
    {
	OutputStream os = null;
	LMReplyPacket reply = null;

	try{
	    os = socket.getOutputStream();
	    reply = new LMReplyPacket(AppMessage.LMREP_GRANTED_SHARED,
				      tokenID,
				      seqNum,
				      objectID);
	    reply.sendSelf(os);
	    os.flush();
	}finally{
	    // NOTE: Do not close any of these, but let gc change ref. counts
	    os = null;
	    socket = null;
	    reply = null;
	}
    }

    //-----------------------------------------------------------------------
    // Send out all shared locks
    //-----------------------------------------------------------------------
    private void distributeSharedLocks(LMLockState lockState)
	throws IOException
    {
	LMRecord record = null;
	String objectID = null;

	objectID = lockState.getObjectID();
	while(!lockState.isSharedQueueEmpty()){
	    record = lockState.dequeueNextShared();
	    lockState.enqueueOutstandingSharedLock(record.getReplySocket(),
						   record.getLockToken(),
						   record.getSeqNum());
	    this.sendSharedLock(record.getReplySocket(),
				record.getLockToken(),
				record.getSeqNum(),
				objectID);
	}
    }

    //-----------------------------------------------------------------------
    // Give out an exclusive lock
    //-----------------------------------------------------------------------
    private void distributeExclusiveLock(LMLockState lockState)
	throws IOException
    {
	LMRecord record = null;
	LMReplyPacket reply = null;
	Socket socket = null;
	OutputStream os = null;

	try{
	    record = lockState.getNextExclusive();
	    reply = new LMReplyPacket(AppMessage.LMREP_GRANTED_EXCLUSIVE,
				      record.getLockToken(),
				      record.getSeqNum(),
				      lockState.getObjectID());
	    socket = record.getReplySocket();
	    os = socket.getOutputStream();
	    reply.sendSelf(os);
	    os.flush();
	}finally{
	    // NOTE: Do not close any of these, but let gc change ref. counts
	    os = null;
	    socket = null;
	    reply = null;
	}
    }

    //-----------------------------------------------------------------------
    // Generate a lock token for a shared lock
    //-----------------------------------------------------------------------
    private int genSharedLockToken()
    {
	// Generate an odd number
	this.lockNum += 2;
	return(lockNum + 1);
    }

    //-----------------------------------------------------------------------
    // Generate a lock token for an exclusive lock
    //-----------------------------------------------------------------------
    private int genExclusiveLockToken()
    {
	// Generate an even number
	this.lockNum += 2;
	return(this.lockNum);
    }

    //-----------------------------------------------------------------------
    // Used for testing
    //-----------------------------------------------------------------------
    public static void main(String[] argv)
    {
	SyncQueue sq = null;

	sq = new SyncQueue();
	new LMHandleThread(sq);
    }

    //-----------------------------------------------------------------------
    // Private data
    //-----------------------------------------------------------------------
    private int lockNum;
    private SyncQueue requestQueue;
    private LMLockStateTable lockStateTable;
}
