import java.net.Socket;
import java.util.LinkedList;
import java.util.Hashtable;
import java.util.Enumeration;

//---------------------------------------------------------------------------
// LMLockState: Used to store all state associated with a lock
//---------------------------------------------------------------------------
public class LMLockState{

    //-----------------------------------------------------------------------
    // Constants
    //-----------------------------------------------------------------------
    public static final int IDLE = 1;
    public static final int EXCLUSIVE = 2;
    public static final int SHARED = 3;

    //-----------------------------------------------------------------------
    // Constructor
    //-----------------------------------------------------------------------
    public LMLockState(String objectID)
    {
	this.objectID = objectID;
	this.stateName = LMLockState.IDLE;
	this.exclusiveQueue = new LinkedList();
	this.sharedQueue = new LinkedList();
	this.outstandingSharedLocks = new Hashtable();
    }

    //-----------------------------------------------------------------------
    // Return the object ID
    //-----------------------------------------------------------------------
    public String getObjectID()
    {
	return(this.objectID);
    }

    //-----------------------------------------------------------------------
    // Return the state name
    //-----------------------------------------------------------------------
    public int getStateName()
    {
	return(this.stateName);
    }

    //-----------------------------------------------------------------------
    // Set the state name
    //-----------------------------------------------------------------------
    public void setStateName(int stateName)
    {
	this.stateName = stateName;
    }

    //-----------------------------------------------------------------------
    // Enqueue a request for an exclusive lock
    //-----------------------------------------------------------------------
    public void enqueueExclusive(Socket replySocket, int tokenID, long seqNum)
    {
	LMRecord record = null;

	record = new LMRecord(replySocket, tokenID, seqNum);
	this.exclusiveQueue.addLast(record);
    }

    //-----------------------------------------------------------------------
    // Enqueue a request for an shared lock
    //-----------------------------------------------------------------------
    public void enqueueShared(Socket replySocket, int tokenID, long seqNum)
    {
	LMRecord record = null;

	record = new LMRecord(replySocket, tokenID, seqNum);
	this.sharedQueue.addLast(record);
    }

    //-----------------------------------------------------------------------
    // Return true if there are no requests for an exclusive lock
    //-----------------------------------------------------------------------
    public boolean isExclusiveQueueEmpty()
    {
	return(this.exclusiveQueue.isEmpty());
    }

    //-----------------------------------------------------------------------
    // Return true if there are no requests for a shared lock
    //-----------------------------------------------------------------------
    public boolean isSharedQueueEmpty()
    {
	return(this.sharedQueue.isEmpty());
    }

    //-----------------------------------------------------------------------
    // Return true if there are no outstanding shared locks
    //-----------------------------------------------------------------------
    public boolean hasNoOutstandingSharedLocks()
    {
	return(this.outstandingSharedLocks.isEmpty());
    }

    //-----------------------------------------------------------------------
    // Remove the head request for an exclusive lock
    //-----------------------------------------------------------------------
    public void dequeueExclusive(int tokenID)
    {
	LMRecord headRecord = null;

	headRecord = (LMRecord)this.exclusiveQueue.removeFirst();
	assert(headRecord.getLockToken() == tokenID);
    }

    //-----------------------------------------------------------------------
    // Add a shared lock request as an outstanding lock request
    //-----------------------------------------------------------------------
    public void enqueueOutstandingSharedLock(Socket replySocket,
					     int tokenID,
					     long seqNum)
    {
	Integer tokenInt = null;
	LMRecord record = null;

	record = new LMRecord(replySocket, tokenID, seqNum);
	tokenInt = new Integer(tokenID);
	assert(this.outstandingSharedLocks.get(tokenInt) == null);
	this.outstandingSharedLocks.put(tokenInt, record);
    }

    //-----------------------------------------------------------------------
    // Remove the shared lock request associated with the given token ID
    //-----------------------------------------------------------------------
    public void dequeueOutstandingSharedLock(int tokenID)
    {
	Integer tokenInt = null;

	tokenInt = new Integer(tokenID);
	assert(this.outstandingSharedLocks.get(tokenInt) != null);
	this.outstandingSharedLocks.remove(tokenInt);
    }

    //-----------------------------------------------------------------------
    // Return the next request for a shared lock
    //-----------------------------------------------------------------------
    public LMRecord dequeueNextShared()
    {
	LMRecord record = null;

	assert(!this.sharedQueue.isEmpty());
	record = (LMRecord)this.sharedQueue.removeFirst();
	return(record);
    }

    //-----------------------------------------------------------------------
    // Return the next request for a exclusive lock
    //-----------------------------------------------------------------------
    public LMRecord getNextExclusive()
    {
	LMRecord record = null;

	assert(!this.exclusiveQueue.isEmpty());
	record = (LMRecord)this.exclusiveQueue.getFirst();
	return(record);
    }

    //-----------------------------------------------------------------------
    // Used for testing
    //-----------------------------------------------------------------------
    public static void main(String[] argv)
    {
	LMLockState ls = null;
	LMRecord lr = null;
	LinkedList l = null;
	int numElements = 0;

	ls = new LMLockState("OID");
	assert(ls.getObjectID().equals("OID"));
	assert(ls.getStateName() == LMLockState.IDLE);
	ls.setStateName(EXCLUSIVE);
	assert(ls.getStateName() == EXCLUSIVE);

	ls.enqueueExclusive(null, 3, 1);
	assert(!ls.isExclusiveQueueEmpty());
	lr = ls.getNextExclusive();
	assert(lr.getLockToken() == 3);
	assert(lr.getSeqNum() == 1);
	ls.dequeueExclusive(3);
	assert(ls.isExclusiveQueueEmpty());

	assert(ls.isSharedQueueEmpty());
	ls.enqueueShared(null, 2, 2);
	assert(!ls.isSharedQueueEmpty());
	ls.enqueueShared(null, 3, 3);
	while(!ls.isSharedQueueEmpty()){
	    lr = ls.dequeueNextShared();
	    assert((lr.getLockToken() == 3) || (lr.getLockToken() == 2));
	    assert((lr.getSeqNum() == 3) || (lr.getSeqNum() == 2));
	    numElements++;
	    ls.enqueueOutstandingSharedLock(lr.getReplySocket(),
					    lr.getLockToken(),
					    lr.getSeqNum());
	}
	assert(numElements == 2);
	ls.dequeueOutstandingSharedLock(3);
	ls.dequeueOutstandingSharedLock(2);
	assert(ls.hasNoOutstandingSharedLocks());
	System.out.println("Success");
    }

    //-----------------------------------------------------------------------
    // Private data
    //-----------------------------------------------------------------------
    private String objectID;
    private int stateName;
    private Hashtable outstandingSharedLocks;
    private LinkedList sharedQueue;
    private LinkedList exclusiveQueue;
}
