//---------------------------------------------------------------------------
// AtomicThread: Implementation of a thread with critical sections
//---------------------------------------------------------------------------
public abstract class AtomicThread implements Runnable{

    //-----------------------------------------------------------------------
    // Constructor
    //-----------------------------------------------------------------------
    public AtomicThread(BooleanLock syncLock)
    {
	this.syncLock = syncLock;
	this.activeFlag = true;
    }

    //-----------------------------------------------------------------------
    // Start this thread
    //-----------------------------------------------------------------------
    public void startSelf()
    {
	Thread thread = null;

	thread = new Thread(this);
	thread.start();
    }

    //-----------------------------------------------------------------------
    // Enable the thread to enter the critical section
    //-----------------------------------------------------------------------
    public synchronized void enable()
    {
	// Will allow thread to get the atomic lock
	this.activeFlag = true;
	this.notify();
    }

    //-----------------------------------------------------------------------
    // Disallow the thread from entering the critical section
    //-----------------------------------------------------------------------
    public synchronized void disable()
    {
	// Will prevent this thread from executing
	this.activeFlag = false;
    }

    //-----------------------------------------------------------------------
    // Return true if the thread is active
    //-----------------------------------------------------------------------
    public synchronized boolean isActive()
    {
	return(this.activeFlag);
    }

    //-----------------------------------------------------------------------
    // Allow a non-atomic thread to enter the critical section
    //-----------------------------------------------------------------------
    public static void acquireCSLock(BooleanLock syncLock)
    {
	try{
	    syncLock.waitToSetTrue(0);
	}catch(InterruptedException e){
	    assert(false);
	}
    }

    //-----------------------------------------------------------------------
    // Allow a non-atomic thread to exit the critical section
    //-----------------------------------------------------------------------
    public static void releaseCSLock(BooleanLock syncLock)
    {
	syncLock.setValue(false);
    }

    //-----------------------------------------------------------------------
    // Allow an instance of AtomicThread to enter the critical section
    //-----------------------------------------------------------------------
    protected void enterCriticalSection()
    {
	boolean active = true;

	do{
	    this.waitUntilActive();
	    try{
		this.syncLock.waitToSetTrue(0);
	    }catch(InterruptedException e){
		assert(false);
	    }
	    // Got the lock; no one can touch me now
	    // Any other thread that wanted to stop this one
	    //  will have to have done this by now. If I am 
	    //  still active, I can stay active as long as I
	    //  want.
	    active = this.isActive();
	    if(!active){
		this.syncLock.setValue(false);
	    }
	    // Note: We use a local "active" variable because if
	    // activeFlag changed here we can get into a state
	    // where we released the lock and exit the method.
	}while(!active);
    }

    //-----------------------------------------------------------------------
    // Allow an instance of AtomicThread to exit the critical section
    //-----------------------------------------------------------------------
    protected void exitCriticalSection()
    {
	assert(this.syncLock.isTrue());
	this.syncLock.setValue(false);
	Thread.yield();
    }

    //-----------------------------------------------------------------------
    // Wait until activeFlag is set to true
    //-----------------------------------------------------------------------
    protected synchronized void waitUntilActive()
    {
	while(!this.activeFlag){
	    try{
		this.wait();
	    }catch(InterruptedException e){
		assert(false);
	    }
	}
    }

    //-----------------------------------------------------------------------
    // Used for testing
    //-----------------------------------------------------------------------
    public static void main(String[] argv)
    {
	BooleanLock syncLock = new BooleanLock(false);
	AtomicThreadTest1 test1 = new AtomicThreadTest1(syncLock);
	AtomicThreadTest2 test2 = new AtomicThreadTest2(syncLock);

	try{
	    Thread.sleep(3000);
	}catch(InterruptedException e){
	    System.err.println("" + e);
	}

	test2.disable();
	try{
	    Thread.sleep(3000);
	}catch(InterruptedException e){
	    System.err.println("" + e);
	}
	test2.enable();
    }

    //-----------------------------------------------------------------------
    // Data members
    //-----------------------------------------------------------------------
    private boolean activeFlag;
    private BooleanLock syncLock;
}

//---------------------------------------------------------------------------
// Used to test AtomicThread
//---------------------------------------------------------------------------
class AtomicThreadTest1 extends AtomicThread{

    //-----------------------------------------------------------------------
    // Constructor
    //-----------------------------------------------------------------------
    public AtomicThreadTest1(BooleanLock syncLock)
    {
	super(syncLock);
	this.startSelf();
    }

    //-----------------------------------------------------------------------
    // Main thread method
    //-----------------------------------------------------------------------
    public void run()
    {
	for(int i = 0; i < 10; i++){
	    this.enterCriticalSection();
	    System.out.println("Something from thread 1");
	    try{
		Thread.sleep(1000);
	    }catch(InterruptedException e){
		// Do nothing
	    }
	    System.out.println("Something else from thread 1");
	    this.exitCriticalSection();
	}
    }
}

//---------------------------------------------------------------------------
// Used to test AtomicThread
//---------------------------------------------------------------------------
class AtomicThreadTest2 extends AtomicThread{

    //-----------------------------------------------------------------------
    // Constructor
    //-----------------------------------------------------------------------
    public AtomicThreadTest2(BooleanLock syncLock)
    {
	super(syncLock);
	this.startSelf();
    }

    //-----------------------------------------------------------------------
    // Main thread method
    //-----------------------------------------------------------------------
    public void run()
    {
	for(int i = 0; i < 10; i++){
	    this.enterCriticalSection();
	    System.out.println("Something from thread 2");
	    try{
		Thread.sleep(1000);
	    }catch(InterruptedException e){
		// Do nothing
	    }
	    System.out.println("Something else from thread 2");
	    this.exitCriticalSection();
	}
    }
}
