//---------------------------------------------------------------------------
// LMReplyPacket: Stores a parsed lock server reply
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
// PACKET FORMAT:
//
// +-------+------+--------+------+-----------+
// | OPNUM | LOCK | SEQNUM | SIZE | OBJECT ID |
// +-------+------+--------+------+-----------+
//
// OPNUM - Bytes 0-3:
//   (See ReplicaImplDoc.txt)
//
// LOCK - Bytes 4-7:
//   Randomly-generated identifier used to represent a lock token
//   0 is used for the global lock token
//
// SEQNUM - Bytes 8-11:
//   Sequence number used to identify callers
//
// SIZE - Bytes 12-15:
//   Size of the object ID (not including trailing 0, if applicable)
//
// OBJECT ID - Bytes 16-?:
//   ID of the object on which to perform the lock operation
//
//---------------------------------------------------------------------------
import java.net.Socket;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.ByteArrayOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;

// Used for testing
import java.io.ByteArrayInputStream;

public class LMReplyPacket implements AppMessage{

    //-----------------------------------------------------------------------
    // Constructor
    //-----------------------------------------------------------------------
    public LMReplyPacket(int result,
			 int lockToken,
			 long seqNum,
			 String objectID)
    {
	this.result = result;
	this.lockToken = lockToken;
	this.seqNum = seqNum;
	this.objectID = objectID;
    }

    //-----------------------------------------------------------------------
    // Constructor
    //-----------------------------------------------------------------------
    public LMReplyPacket(InputStream is) throws IOException
    {
	int size = 0;
	byte[] objectIDBytes = null;
	DataInputStream dis = null;

	try{
	    dis = new DataInputStream(is);
	    this.result = dis.readInt();
	    this.lockToken = dis.readInt();
	    this.seqNum = dis.readLong();
	    size = LMReplyPacket.readInt(is);
	    objectIDBytes = new byte[size];
	    dis.readFully(objectIDBytes);
	    this.objectID = new String(objectIDBytes);
	}finally{
	    dis = null;
	    objectIDBytes = null;
	}
    }

    //-----------------------------------------------------------------------
    // Serialize the reply packet onto the output stream
    //-----------------------------------------------------------------------
    public void sendSelf(OutputStream os) throws IOException
    {
	byte[] objectIDBytes = null;
	ByteArrayOutputStream baos = null;
	byte[] dataBytes = null;
	DataOutputStream dos = null;

	try{
	    baos = new ByteArrayOutputStream(1000);
	    dos = new DataOutputStream(baos);
	    dos.writeInt(this.result);
	    dos.writeInt(this.lockToken);
	    dos.writeLong(this.seqNum);
	    objectIDBytes = this.objectID.getBytes();
	    dos.writeInt(objectIDBytes.length);
	    dos.write(objectIDBytes);
	    dos.flush();
	    dataBytes = baos.toByteArray();
	    os.write(dataBytes);
	}finally{
	    baos = null;
	    dos = null;
	    objectIDBytes = null;
	    dataBytes = null;
	}
    }

    //-----------------------------------------------------------------------
    // Return the sequence number (not applicable)
    //-----------------------------------------------------------------------
    public long getSeqNum()
    {
	return(this.seqNum);
    }

    //-----------------------------------------------------------------------
    // Return the result
    //-----------------------------------------------------------------------
    public int getMessageType()
    {
	return(this.result);
    }

    //-----------------------------------------------------------------------
    // Return the objectID
    //-----------------------------------------------------------------------
    public String getObjectID()
    {
	return(this.objectID);
    }

    //-----------------------------------------------------------------------
    // Return true if the message contains a lock
    //-----------------------------------------------------------------------
    public boolean containsLock()
    {
	return(true);
    }

    //-----------------------------------------------------------------------
    // Return the lock token
    //-----------------------------------------------------------------------
    public int getLockToken()
    {
	assert(this.lockToken >= 0);
	return(this.lockToken);
    }

    //-----------------------------------------------------------------------
    // Return any data present in the message (not applicable)
    //-----------------------------------------------------------------------
    public byte[] getData()
    {
	assert(false);
	return(null);
    }

    //-----------------------------------------------------------------------
    // Return the reply socket (not applicable)
    //-----------------------------------------------------------------------
    public Socket getReplySocket()
    {
	assert(false);
	return(null);
    }

    //-----------------------------------------------------------------------
    // Return the timestamp (not applicable)
    //-----------------------------------------------------------------------
    public long getTimeStamp()
    {
	assert(false);
	return(0);
    }

    //-----------------------------------------------------------------------
    // Convert to a string
    //-----------------------------------------------------------------------
    public String toString()
    {
	String retStr = null;

	retStr = "(";
	switch(this.result){
	case AppMessage.LMREP_GRANTED_SHARED:
	    retStr += "LMREP_LOCK_GRANTED_SHARED";
	    break;
	case AppMessage.LMREP_GRANTED_EXCLUSIVE:
	    retStr += "LMREP_LOCK_GRANTED_EXCLUSIVE";
	    break;
	default:
	    retStr += "" + this.result;
	    break;
	}
	retStr += ", " + this.lockToken;
	retStr += ", " + this.objectID.length();
	retStr += ", \"" + this.objectID + "\")";
	return(retStr);
    }

    //-----------------------------------------------------------------------
    // Read an integer from a stream (4 bytes of data)
    //-----------------------------------------------------------------------
    private static int readInt(InputStream is) throws IOException
    {
        int num = 0;
        int numRead = 0;
        byte[] intHolder = null;

        intHolder = new byte[4];
        numRead = is.read(intHolder);
        while(numRead < 4){
            numRead += is.read(intHolder, numRead, 4 - numRead);
        }

        num = intHolder[0] & 0xFF;
        num = (num << 8) | (intHolder[1] & 0xFF);
        num = (num << 8) | (intHolder[2] & 0xFF);
        num = (num << 8) | (intHolder[3] & 0xFF);
        return(num);
    }

    //-----------------------------------------------------------------------
    // Write an integer to a stream (4 bytes of data)
    //-----------------------------------------------------------------------
    private static void writeInt(int value, OutputStream os)
	throws IOException
    {
	byte[] intHolder = null;

	intHolder = new byte[4];
	intHolder[3] = (byte)(value & 0xFF);
	intHolder[2] = (byte)((value >> 8) & 0xFF);
	intHolder[1] = (byte)((value >> 16) & 0xFF);
	intHolder[0] = (byte)((value >> 24) & 0xFF);
	os.write(intHolder);
    }

    //-----------------------------------------------------------------------
    // Used for testing only
    //-----------------------------------------------------------------------
    public static void main(String[] argv)
    {
	byte[] b = null;
	byte[] b2 = null;
	String x = null;
	String str = null;
	LMReplyPacket rp = null;
	LMReplyPacket rp2 = null;
	ByteArrayInputStream bais = null;
	ByteArrayOutputStream baos = null;

	try{
	    x = new String("Something random");
	    rp = new LMReplyPacket(AppMessage.LMREP_GRANTED_SHARED, 2, 1, x);
	    baos = new ByteArrayOutputStream();
	    rp.sendSelf(baos);
	    b = baos.toByteArray();
	    b2 = baos.toByteArray();
	    bais = new ByteArrayInputStream(b2);
	    rp2 = new LMReplyPacket(bais);
	    System.out.println("From self: " + rp2);
	    for(int i = 0; i < 12; i++){
		System.out.println("" + b[i]);
	    }
	    str = new String(b, 12, b.length - 12);
	    System.out.println(str);
	    System.out.println("getMessageType() = " + rp.getMessageType());
	    System.out.println("getObjectID() = \"" +
			       new String(rp.getObjectID()) + "\"");
	    System.out.println("containsLock() = " + rp.containsLock());
	    System.out.println("getLockToken() = " + rp.getLockToken());
	}catch(IOException e){
	    System.err.println("" + e);
	}
    }

    //-----------------------------------------------------------------------
    // Private data
    //-----------------------------------------------------------------------
    private int result;
    private int lockToken;
    private long seqNum;
    private String objectID;
}
