//---------------------------------------------------------------------------
// LMRequestPacket: Stores a parsed client request
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
// PACKET FORMAT:
//
// +-------+------+--------+------+-----------+
// | OPNUM | LOCK | SEQNUM | SIZE | OBJECT ID |
// +-------+------+--------+------+-----------+
//
// OPNUM - Bytes 0-3:
//   (See ReplicaImplDoc.txt)
//
// LOCK - Bytes 4-7:
//   Randomly-generated identifier used to represent a lock token
//   0 is used for the global lock token
//
// SEQNUM - Bytes 8-11:
//   Sequence number used to identify callers
//
// SIZE - Bytes 12-15:
//   Size of the object ID (not including trailing 0, if applicable)
//
// OBJECT ID - Bytes 16-?:
//   ID of the object on which to perform the lock operation
//
//---------------------------------------------------------------------------
import java.net.Socket;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.ByteArrayOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;

// Used for testing
import java.io.ByteArrayInputStream;

public class LMRequestPacket implements AppMessage{

    //-----------------------------------------------------------------------
    // Constructor
    //-----------------------------------------------------------------------
    public LMRequestPacket(InputStream is, Socket replySocket)
	throws IOException
    {
	int size = 0;
	DataInputStream dis = null;
	byte[] objectIDBytes = null;

	try{
	    dis = new DataInputStream(is);
	    this.operation = dis.readInt();
	    this.lockToken = dis.readInt();
	    this.seqNum = dis.readLong();
	    size = dis.readInt();
	    objectIDBytes = new byte[size];
	    dis.readFully(objectIDBytes);
	    this.objectID = new String(objectIDBytes);
	    this.replySocket = replySocket;
	    assert((this.operation >= AppMessage.LMOP_ACQUIRE_EXCLUSIVE) &&
		   (this.operation <= AppMessage.LMOP_RELEASE));
	}finally{
	    dis = null;
	    objectIDBytes = null;
	}
    }

    //-----------------------------------------------------------------------
    // Constructor
    //-----------------------------------------------------------------------
    public LMRequestPacket(int operation,
			   int lockToken,
			   long seqNum,
			   String objectID,
			   Socket replySocket)
    {
	this.operation = operation;
	this.lockToken = lockToken;
	this.seqNum = seqNum;
	this.objectID = objectID;
	this.replySocket = replySocket;
	assert((this.operation >= AppMessage.LMOP_ACQUIRE_EXCLUSIVE) &&
	       (this.operation <= AppMessage.LMOP_RELEASE));
    }

    //-----------------------------------------------------------------------
    // Serlialize the request packet over the output stream
    //-----------------------------------------------------------------------
    public void sendSelf(OutputStream os) throws IOException
    {
	byte[] objectIDBytes = null;
	ByteArrayOutputStream baos = null;
	byte[] dataBytes = null;
	DataOutputStream dos = null;

	try{
	    baos = new ByteArrayOutputStream(1000);
	    baos.reset();
	    dos = new DataOutputStream(baos);
	    dos.writeInt(this.operation);
	    dos.writeInt(this.lockToken);
	    dos.writeLong(this.seqNum);
	    objectIDBytes = this.objectID.getBytes();
	    dos.writeInt(objectIDBytes.length);
	    dos.write(objectIDBytes);
	    dos.flush();
	    dataBytes = baos.toByteArray();
	    os.write(dataBytes);
	}finally{
	    baos = null;
	    dos = null;
	    objectIDBytes = null;
	    dataBytes = null;
	}
    }

    //-----------------------------------------------------------------------
    // Return the sequence number (not applicable)
    //-----------------------------------------------------------------------
    public long getSeqNum()
    {
	return(this.seqNum);
    }

    //-----------------------------------------------------------------------
    // Return the operation
    //-----------------------------------------------------------------------
    public int getMessageType()
    {
	return(this.operation);
    }

    //-----------------------------------------------------------------------
    // Return the objectID
    //-----------------------------------------------------------------------
    public String getObjectID()
    {
	return(this.objectID);
    }

    //-----------------------------------------------------------------------
    // Return true if the request contains a lock
    //-----------------------------------------------------------------------
    public boolean containsLock()
    {
	boolean hasLock = false;

	hasLock = (this.operation == AppMessage.LMOP_UPGRADE) ||
	    (this.operation == AppMessage.LMOP_DOWNGRADE) ||
	    (this.operation == AppMessage.LMOP_RELEASE);
	return(hasLock);
    }

    //-----------------------------------------------------------------------
    // Return the lock token
    //-----------------------------------------------------------------------
    public int getLockToken()
    {
	assert(this.containsLock());
	return(this.lockToken);
    }

    //-----------------------------------------------------------------------
    // Return any data present in the message (not applicable)
    //-----------------------------------------------------------------------
    public byte[] getData()
    {
	assert(false);
	return(null);
    }

    //-----------------------------------------------------------------------
    // Return the reply socket associated with this message
    //-----------------------------------------------------------------------
    public Socket getReplySocket()
    {
	return(this.replySocket);
    }

    //-----------------------------------------------------------------------
    // Return the timestamp (not applicable)
    //-----------------------------------------------------------------------
    public long getTimeStamp()
    {
	assert(false);
	return(0);
    }

    //-----------------------------------------------------------------------
    // Convert to a string
    //-----------------------------------------------------------------------
    public String toString()
    {
	String retStr = null;

	retStr = "(";
	switch(this.operation){
	case AppMessage.LMOP_ACQUIRE_EXCLUSIVE:
	    retStr += "LMOP_ACQUIRE_EXCLUSIVE";
	    break;
	case AppMessage.LMOP_ACQUIRE_SHARED:
	    retStr += "LMOP_ACQUIRE_SHARED";
	    break;
	case AppMessage.LMOP_UPGRADE:
	    retStr += "LMOP_UPGRADE";
	    break;
	case AppMessage.LMOP_DOWNGRADE:
	    retStr += "LMOP_DOWNGRADE";
	    break;
	case AppMessage.LMOP_RELEASE:
	    retStr += "LMOP_RELEASE";
	    break;
	default:
	    retStr += "" + this.operation;
	    break;
	}
	retStr += ", " + this.lockToken;
	retStr += ", " + this.objectID.length();
	retStr += ", \"" + this.objectID + "\")";
	return(retStr);
    }

    //-----------------------------------------------------------------------
    // Used for testing only
    //-----------------------------------------------------------------------
    public static void main(String[] argv)
    {
	byte[] b = null;
	byte[] x = null;
	LMRequestPacket rp = null;
	ByteArrayInputStream bais = null;
	ByteArrayOutputStream baos = null;

	try{
	    x = (new String("Something random")).getBytes();
	    b = new byte[12 + x.length];
	    b[0] = 0;
	    b[1] = 0;
	    b[2] = 0;
	    b[3] = AppMessage.LMOP_UPGRADE;
	    b[4] = 0;
	    b[5] = 0;
	    b[6] = 0;
	    b[7] = 2;
	    b[8] = 0;
	    b[9] = 0;
	    b[10] = 0;
	    b[11] = (byte)x.length;
	    for(int i = 0; i < x.length; i++){
		b[12 + i] = x[i];
	    }
	    bais = new ByteArrayInputStream(b);
	    rp = new LMRequestPacket(bais, null);
	    System.out.println("" + rp);
	    System.out.println("getMessageType = " + rp.getMessageType());
	    System.out.println("getLockToken = " + rp.getLockToken());
	    System.out.println("getObjectID = \"" + rp.getObjectID() + "\"");
	    System.out.println("containsLock = " + rp.containsLock());
	    System.out.println("getReplySocket = " + rp.getReplySocket());

	    baos = new ByteArrayOutputStream();
	    rp.sendSelf(baos);
	    b = baos.toByteArray();
	    bais = new ByteArrayInputStream(b);
	    rp = new LMRequestPacket(bais, null);
	    System.out.println("From self: " + rp);
	}catch(IOException e){
	    System.err.println("" + e);
	}
    }

    //-----------------------------------------------------------------------
    // Private data
    //-----------------------------------------------------------------------
    private int operation;
    private int lockToken;
    private long seqNum;
    private String objectID;
    private Socket replySocket;
}
