//---------------------------------------------------------------------------
// LocalFS: Interface to the local file system on the replica
//---------------------------------------------------------------------------
import java.io.File;
import java.io.IOException;
import java.io.FileOutputStream;

public class LocalFS{

    //-----------------------------------------------------------------------
    // Constructor
    //-----------------------------------------------------------------------
    public LocalFS()
    {
	this.shadowFileMgr = new ShadowFileMgr();
    }

    //-----------------------------------------------------------------------
    // Update an object on disk
    //-----------------------------------------------------------------------
    public void updateDisk(String objectID,
			   byte[] data,
			   int lockToken,
			   long seqNum) throws IOException
    {
	// Write data first, then write the shadow file (fault-tolerance)
	FileOutputStream fos = null;
	ShadowEntry entry = null;

	assert(this.isExclusiveLock(lockToken));
	this.createParentDirs(objectID);
	fos = new FileOutputStream(objectID);
	fos.write(data);
	fos.close();
	entry = new ShadowEntry(objectID,
				seqNum,
				ShadowFileMgr.STATUS_DATA);
	this.shadowFileMgr.saveEntry(entry);
    }

    //-----------------------------------------------------------------------
    // Invalidate an object on disk
    //-----------------------------------------------------------------------
    public void invalidateDisk(String objectID, int lockToken, long seqNum)
	throws IOException
    {
	ShadowEntry entry = null;

	assert(this.isExclusiveLock(lockToken));
	this.createParentDirs(objectID);
	entry = new ShadowEntry(objectID,
				seqNum,
				ShadowFileMgr.STATUS_INV);
	this.shadowFileMgr.saveEntry(entry);
    }

    //-----------------------------------------------------------------------
    // Return the sequence number of the object on disk
    //-----------------------------------------------------------------------
    public long getSeqNum(String objectID, int lockToken)
    {
	ShadowEntry entry = null;

	assert(this.isExclusiveLock(lockToken));
	entry = this.shadowFileMgr.getEntry(objectID);
	return(entry.getSeqNum());
    }

    //-----------------------------------------------------------------------
    // Used for testing only
    //-----------------------------------------------------------------------
    public static void main(String[] argv)
    {
	LocalFS localFS = null;

	try{
	    localFS = new LocalFS();
	    localFS.updateDisk("a",
			       (new String("something")).getBytes(),
			       2,
			       22);
	    System.out.println("SeqNum = " + localFS.getSeqNum("a", 2));
	    localFS.invalidateDisk("a", 2, 21);
	    System.out.println("SeqNum = " + localFS.getSeqNum("a", 2));
	}catch(IOException e){
	    System.err.println("" + e);
	}
    }

    //-----------------------------------------------------------------------
    // Return true if the passed-in lock is an exclusive lock
    //-----------------------------------------------------------------------
    private boolean isExclusiveLock(int lockToken)
    {
	return((lockToken != 0) && ((lockToken % 2) == 0));
    }

    //-----------------------------------------------------------------------
    // Create all parent directories of this object
    //-----------------------------------------------------------------------
    private void createParentDirs(String objectID)
    {
	int lastSlashIndex = 0;
	boolean success = false;
	File parentDirFile = null;
	String parentDirString = null;

	// Extract the parent directory name
	lastSlashIndex = objectID.lastIndexOf('/');
	if(lastSlashIndex >= 0){
	    parentDirString = objectID.substring(0, lastSlashIndex);
	    parentDirFile = new File(parentDirString);
	    if(!parentDirFile.exists()){
		success = parentDirFile.mkdirs();
		assert(success);
	    }else{
		// Don't need to do anything.
	    }
	}else{
	    // Assume relative path name and leave as is.
	}
    }

    //-----------------------------------------------------------------------
    // Protected data
    //-----------------------------------------------------------------------
    protected ShadowFileMgr shadowFileMgr;
}
