import java.net.Socket;
import java.io.IOException;
import java.io.InputStream;
import java.io.DataInputStream;

// For testing
import java.net.InetAddress;
import java.net.ServerSocket;
import java.io.OutputStream;
import java.io.DataOutputStream;
import java.util.Date;

//---------------------------------------------------------------------------
// MQRecvThread: Implements a thread that receives messages
//---------------------------------------------------------------------------
public class MQRecvThread implements Runnable{

    //-----------------------------------------------------------------------
    // Constructor
    //-----------------------------------------------------------------------
    public MQRecvThread(SyncQueue localQueue)
    {
	this.localQueue = localQueue;
	this.activeFlag = false;
	this.socket = null;
	this.dataIS = null;
	(new Thread(this)).start();
    }

    //-----------------------------------------------------------------------
    // Main thread method
    //-----------------------------------------------------------------------
    public void run()
    {
	OriginMsg msg = null;

	try{
	    while(true){
		this.waitUntilActive();
		assert(this.socket != null);
		assert(this.dataIS != null);
		try{
		    msg = new OriginMsg(this.dataIS);
		    this.localQueue.enqueue(msg);
		    msg = null;
		}catch(IOException e){
		    this.stopThread();
		}
	    }
	}catch(Throwable t){
	    t.printStackTrace();
	    assert(false);
	}
    }

    //-----------------------------------------------------------------------
    // Start the thread
    //-----------------------------------------------------------------------
    public synchronized void startThread(Socket socket) throws IOException
    {
	InputStream is = null;

	this.closeSocket();
	this.socket = socket;
	is = socket.getInputStream();
	this.dataIS = new DataInputStream(is);
	this.activeFlag = true;
	this.notifyAll();
    }

    //-----------------------------------------------------------------------
    // Stop the thread and do some cleanup
    //-----------------------------------------------------------------------
    public synchronized void stopThread()
    {
	this.closeSocket();
	this.activeFlag = false;
    }

    //-----------------------------------------------------------------------
    // Cause the thread to block until it is woken up again
    //-----------------------------------------------------------------------
    private synchronized void waitUntilActive()
    {
	while(!this.activeFlag){
	    try{
		this.wait();
	    }catch(InterruptedException e){
		System.err.println("" + e);
	    }
	}
    }

    //-----------------------------------------------------------------------
    // Do some cleanup by closing the socket and any associated input streams
    //-----------------------------------------------------------------------
    private void closeSocket()
    {
	InputStream is = null;

	if(this.dataIS != null){
	    try{
		this.dataIS.close();
	    }catch(IOException e){
		System.err.println("MQRecvThread (closeSocket): " + e);
	    }
	}
	if(this.socket != null){
	    try{
		is = this.socket.getInputStream();
		is.close();
	    }catch(IOException e){
		System.err.println("MQRecvThread (closeSocket): " + e);
	    }
	    try{
		this.socket.close();
	    }catch(IOException e){
		System.err.println("MQRecvThread (closeSocket): " + e);
	    }
	}
	this.dataIS = null;
	this.socket = null;
    }

    //-----------------------------------------------------------------------
    // Used for testing
    //-----------------------------------------------------------------------
    public static void main(String[] argv)
    {
	OriginMsg msg = null;
	Socket socket = null;
	SyncQueue queue = null;
	MQRecvThread recvThread = null;

	try{
	    new RecvThreadServer();
	    queue = new SyncQueue();
	    recvThread = new MQRecvThread(queue);

	    socket = new Socket(InetAddress.getLocalHost(), 4555);
	    socket.setTcpNoDelay(true);
	    socket.setReceiveBufferSize(Globals.TCP_BUFFER_SIZE);
	    System.out.println("About to start receive thread");
	    recvThread.startThread(socket);
	    msg = (OriginMsg)queue.peekBlocked();
	    queue.dequeue();
	    System.out.println("Received: " + (msg.getData())[0]);

	    socket = new Socket(InetAddress.getLocalHost(), 4555);
	    socket.setTcpNoDelay(true);
	    socket.setReceiveBufferSize(Globals.TCP_BUFFER_SIZE);
	    System.out.println("About to start receive thread");
	    recvThread.startThread(socket);
	    msg = (OriginMsg)queue.peekBlocked();
	    queue.dequeue();
	    System.out.println("Received: " + (msg.getData())[0]);
	}catch(IOException e){
	    System.err.println("" + e);
	    Thread.dumpStack();
	}
    }

    //-----------------------------------------------------------------------
    // Private data
    //-----------------------------------------------------------------------
    Socket socket;
    boolean activeFlag;
    SyncQueue localQueue;
    DataInputStream dataIS;
}

//---------------------------------------------------------------------------
// RecvThreadServer: Implement a fake server used for testing
//---------------------------------------------------------------------------
class RecvThreadServer extends Thread{

    //-----------------------------------------------------------------------
    // Constructor
    //-----------------------------------------------------------------------
    public RecvThreadServer()
    {
	this.start();
    }

    //-----------------------------------------------------------------------
    // Main thread method
    //-----------------------------------------------------------------------
    public void run()
    {
	Socket socket = null;
	ServerSocket serverSocket = null;
	OutputStream os = null;
	DataOutputStream dos = null;
	int num = 3;
	byte[] b = null;

	try{
	    serverSocket = new ServerSocket(4555);
	    b = new byte[1];
	    while(true){
		socket = serverSocket.accept();
		socket.setTcpNoDelay(true);
		os = socket.getOutputStream();
		dos = new DataOutputStream(os);
		System.out.println("About to write object with num = " + num);
		b[0] = (byte)num++;
		(new OriginMsg(0, AppMessage.READ, null, b)).sendSelf(dos);
		dos.close();
		os.close();
		socket.close();
	    }
	}catch(IOException e){
	    System.err.println("" + e);
	}
    }
}
