import java.net.Socket;
import java.net.InetAddress;
import java.io.IOException;
import java.io.OutputStream;
import java.io.DataOutputStream;
import java.io.ByteArrayOutputStream;

// For testing
import java.net.ServerSocket;
import java.io.InputStream;
import java.io.DataInputStream;

//---------------------------------------------------------------------------
// MQSendThread: Implements a thread that continuously sends messages
//---------------------------------------------------------------------------
public class MQSendThread implements Runnable{

    //-----------------------------------------------------------------------
    // Constants
    //-----------------------------------------------------------------------
    public static final long INIT_TIMEOUT = 100;
    public static final long MAX_TIMEOUT = 10000;
    public static final double TIMEOUT_MULT_FACTOR = 2.0;

    //-----------------------------------------------------------------------
    // Constructor
    //-----------------------------------------------------------------------
    public MQSendThread(SyncQueue localQueue,
			InetAddress address,
			int portNum) throws IOException
    {
	this.localQueue = localQueue;
	this.address = address;
	this.portNum = portNum;
	(new Thread(this)).start();
    }

    //-----------------------------------------------------------------------
    // Main thread method
    //-----------------------------------------------------------------------
    public void run()
    {
	OriginMsg msg = null;
	Socket socket = null;
	OutputStream os = null;
	byte[] dataBytes = null;
	DataOutputStream dataOS = null;
	ByteArrayOutputStream baos = null;

	try{
	    while(true){
		socket = this.connect();
		try{
		    os = socket.getOutputStream();
		    baos = new ByteArrayOutputStream(1000);
		    dataOS = new DataOutputStream(baos);
		    while(true){
			baos.reset();
			msg = (OriginMsg)this.localQueue.peekBlocked();
			msg.sendSelf(dataOS);
			dataOS.flush();
			dataBytes = baos.toByteArray();
			os.write(dataBytes);
			os.flush();
			this.localQueue.dequeue();
			Globals.stats.incRegBW(dataBytes.length);
			Globals.stats.incRegDataBW(msg);
		    }
		}catch(IOException e){
		    System.err.println("MQSendThread (run 1): ");
		    if(dataOS != null){
			try{
			    dataOS.close();
			}catch(IOException e1){
			    System.err.println("MQSendThread (run 2): " + e1);
			}
		    }
		    this.closeSocket(socket);
		}
	    }
	}catch(Throwable t){
	    t.printStackTrace();
	    assert(false);
	}
    }

    //-----------------------------------------------------------------------
    // Keep trying to connect until a connection is estabilished
    //-----------------------------------------------------------------------
    private Socket connect()
    {
	long timeout = 0L;
	long wakeTime = 0L;
	Socket socket = null;
	boolean connected = false;

	timeout = MQSendThread.INIT_TIMEOUT;
	connected = false;
	while(!connected){
	    try{
		socket = new Socket(this.address, this.portNum);
		socket.setTcpNoDelay(true);
		socket.setSendBufferSize(Globals.TCP_BUFFER_SIZE);
		connected = true;
	    }catch(IOException e){
		System.err.println("MQSendThread (connect 1)" + e);
		wakeTime = System.currentTimeMillis() + timeout;
		while(wakeTime > System.currentTimeMillis()){
		    try{
			Thread.sleep(wakeTime - System.currentTimeMillis());
		    }catch(InterruptedException ie){
			System.err.println("" + ie);
		    }
		}
		timeout = (long)(timeout * MQSendThread.TIMEOUT_MULT_FACTOR);
		if(timeout > MQSendThread.MAX_TIMEOUT){
		    timeout = MQSendThread.MAX_TIMEOUT;
		}
	    }
	}
	return(socket);
    }

    //-----------------------------------------------------------------------
    // Close a socket and its output stream
    //-----------------------------------------------------------------------
    private void closeSocket(Socket socket)
    {
	OutputStream os = null;

	assert(socket != null);
	try{
	    os = socket.getOutputStream();
	    os.close();
	}catch(IOException e){
	    // Do nothing
	}
	try{
	    socket.close();
	}catch(IOException e){
	    // Do nothing
	}
    }

    //-----------------------------------------------------------------------
    // Used for testing
    //-----------------------------------------------------------------------
    public static void main(String[] argv)
    {
	SyncQueue queue1 = null;
	SyncQueue queue2 = null;
	int num = 1;
	byte[] b = null;

	try{
	    queue1 = new SyncQueue();
	    queue2 = new SyncQueue();
	    new MQSendThread(queue1,
			     InetAddress.getLocalHost(),
			     4555);
	    //new MQSendThread(queue2, InetAddress.getLocalHost());
	    //Thread.sleep(2000);
	    new SendThreadServer();
	    b = new byte[1];
	    for(int i = 0; i < 5; i++){
		Thread.sleep(1000);
		b[0] = (byte)num++;
		queue1.enqueue(new OriginMsg(0, AppMessage.READ, null, b));
		//queue2.enqueue(new OriginMsg(4));
	    }
	}catch(IOException e){
	    System.err.println("" + e);
	}catch(InterruptedException e){
	    System.err.println("" + e);
	}
    }

    //-----------------------------------------------------------------------
    // Private data
    //-----------------------------------------------------------------------
    protected int portNum;
    protected InetAddress address;
    protected SyncQueue localQueue;
}

//---------------------------------------------------------------------------
// SendThreadServer: Emulate a server to test MQSendThread
//---------------------------------------------------------------------------
class SendThreadServer extends Thread{

    //-----------------------------------------------------------------------
    // Constructor
    //-----------------------------------------------------------------------
    public SendThreadServer()
    {
	this.active = true;
	this.start();
    }

    //-----------------------------------------------------------------------
    // Stop the thread
    //-----------------------------------------------------------------------
    public synchronized void stopThread()
    {
	this.active = false;
    }

    //-----------------------------------------------------------------------
    // Return true if the thread is active
    //-----------------------------------------------------------------------
    private synchronized boolean isActive()
    {
	return(this.active);
    }

    //-----------------------------------------------------------------------
    // Main thread method
    //-----------------------------------------------------------------------
    public void run()
    {
	int num = 0;
	Socket socket = null;
	ServerSocket serverSocket = null;

	try{
	    serverSocket = new ServerSocket(4555);
	    while(this.isActive()){
		socket = serverSocket.accept();
		socket.setTcpNoDelay(true);
		new SendThreadServerHandler(socket, num++);
	    }
	}catch(IOException e){
	    System.err.println("" + e);
	}
    }

    //-----------------------------------------------------------------------
    // Private data
    //-----------------------------------------------------------------------
    private boolean active;
}

//---------------------------------------------------------------------------
// SendThreadServerHandler: Handle requests
//---------------------------------------------------------------------------
class SendThreadServerHandler extends Thread{

    //-----------------------------------------------------------------------
    // Constructor
    //-----------------------------------------------------------------------
    public SendThreadServerHandler(Socket socket, int threadNum)
    {
	this.socket = socket;
	this.threadNum = threadNum;
	this.start();
    }

    //-----------------------------------------------------------------------
    // Main thread method
    //-----------------------------------------------------------------------
    public void run()
    {
	OriginMsg msg = null;
	InputStream is = null;
	DataInputStream dis = null;
	int num = 0;

	try{
	    is = this.socket.getInputStream();
	    dis = new DataInputStream(is);
	    while(num < 2){
		msg = new OriginMsg(dis);
		System.out.println("" + this.threadNum + ": " +
				   (msg.getData())[0]);
		num++;
	    }
	    dis.close();
	    is.close();
	    this.socket.close();
	}catch(IOException e){
	    System.err.println("" + e);
	}
    }

    //-----------------------------------------------------------------------
    // Private data
    //-----------------------------------------------------------------------
    protected Socket socket;
    protected int threadNum;
}
