//---------------------------------------------------------------------------
// OriginEmulator: Emulates updates at the origin site
//---------------------------------------------------------------------------
import java.io.File;
import java.io.IOException;
import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.io.FileOutputStream;

public class OriginEmulator{

    //-----------------------------------------------------------------------
    // Constants
    //-----------------------------------------------------------------------
    public static final String OLD_PREFIX = "/jfs/daedalus/cddo/sink";
    public static final int OLD_PREFIX_LEN = 23;
    //public static final String NEW_PREFIX = "/home/root/root_dir";
    public static final String NEW_PREFIX = "/var/local/root_dir";

    //-----------------------------------------------------------------------
    // Constructor
    //-----------------------------------------------------------------------
    public OriginEmulator(String[] replicaNames,
			  String priorityFileName,
			  int algorithm,
			  int regPort,
			  int nicePort,
			  int ackPort)
    {
	InputStreamReader isr = null;

	assert((algorithm >= Origin.DEMAND_ONLY_ALGO) &&
	       (algorithm <= Origin.PREF_INVAL_ALGO));
	this.lmComm = new SyncLMCommPoint(new LMCommPoint());
	this.origin = new Origin(algorithm,
				 lmComm,
				 replicaNames,
				 regPort,
				 nicePort,
				 ackPort);
	this.offsetMS = 0;

	try{
	    this.priorityTracker = new PriorityTracker(priorityFileName);
	}catch(IOException e){
	    System.err.println("OriginEmulator (OriginEmulator 1): " + e);
	}
	isr = new InputStreamReader(System.in);
	this.stdinReader = new BufferedReader(isr);
    }

    //-----------------------------------------------------------------------
    // Start the emulate loop
    //-----------------------------------------------------------------------
    public void enterEmulateLoop() throws IOException
    {
	UpdateTraceRecord nextUR = null;
	long nextUpdateTimeMS = 0L;
	AppMessage lockMsg = null;
	int lockToken = 0;
	double priority = 0;

	nextUR = this.readNextUpdateRecord();
	if(nextUR != null){
	    this.offsetMS = System.currentTimeMillis() -
		nextUR.getUpdateTime();
	}

	while(nextUR != null){
	    nextUpdateTimeMS = this.offsetMS + nextUR.getUpdateTime();
	    while(System.currentTimeMillis() < nextUpdateTimeMS){
		try{
		    Thread.sleep(nextUpdateTimeMS -
				 System.currentTimeMillis());
		}catch(InterruptedException e){
		    // Do nothing
		}
	    }
	    lockMsg = this.lmComm.requestExclusiveLock(nextUR.getObjectID());
	    assert(lockMsg.getMessageType() ==
		   AppMessage.LMREP_GRANTED_EXCLUSIVE);
	    lockToken = lockMsg.getLockToken();
	    this.updateLocalDisk(nextUR.getObjectID(),
				 nextUR.getObjectSize(),
				 lockToken);
	    priority = this.priorityTracker.getPriority(nextUR.objectID);
	    this.origin.replicateData(nextUR.getObjectID(),
				      lockToken,
				      priority);
	    this.lmComm.releaseLock(nextUR.getObjectID(), lockToken);
	    nextUR = this.readNextUpdateRecord();
	}
    }

    //-----------------------------------------------------------------------
    // Start the emulate loop
    //-----------------------------------------------------------------------
    private UpdateTraceRecord readNextUpdateRecord() throws IOException
    {
	String line = null;
	String objectID = null;
	UpdateTraceRecord utr = null;
	UpdateTraceRecord localDiskRecord = null;

	line = this.stdinReader.readLine();
	if(line != null){
	    utr = new UpdateTraceRecord(line);
	    objectID = utr.getObjectID();
	    assert(objectID.startsWith(OLD_PREFIX));
	    objectID = NEW_PREFIX + objectID.substring(OLD_PREFIX_LEN);
	    localDiskRecord = new UpdateTraceRecord(objectID,
						    utr.getObjectSize(),
						    utr.getUpdateTime());
	    utr = null;
	}
	return(localDiskRecord);
    }

    //-----------------------------------------------------------------------
    // Write a (possibly new) file to disk
    //-----------------------------------------------------------------------
    public static void updateLocalDisk(String objectID,
				       long objectSize,
				       int lockToken)
	throws IOException
    {
	OriginEmulator.genFile(objectID, lockToken);
	OriginEmulator.fillFile(objectID, objectSize, lockToken);
    }

    //-----------------------------------------------------------------------
    // Origin starting point
    //-----------------------------------------------------------------------
    public static void main(String[] argv)
    {
	int regPort = 0;
	int nicePort = 0;
	int ackPort = 0;
	int algorithm = 0;
	int sleepPort = 0;
	String[] replicaNames = null;
	String priorityFileName = null;
	OriginEmulator originEmulator = null;

	try{
	    if(argv.length > 6){
		replicaNames = new String[argv.length - 6];
		System.arraycopy(argv, 0, replicaNames, 0, argv.length - 6);
		priorityFileName = argv[argv.length - 6];
		algorithm = Integer.parseInt(argv[argv.length - 5]);
		regPort = Integer.parseInt(argv[argv.length - 4]);
		nicePort = Integer.parseInt(argv[argv.length - 3]);
		ackPort = Integer.parseInt(argv[argv.length - 2]);
		sleepPort = Integer.parseInt(argv[argv.length - 1]);
		TestbedWait.waitForTestStart(sleepPort);
		originEmulator = new OriginEmulator(replicaNames,
						    priorityFileName,
						    algorithm,
						    regPort,
						    nicePort,
						    ackPort);
		originEmulator.enterEmulateLoop();
	    }else{
		System.err.println("Usage: OriginEmulator <replicanames>+ " +
				   "<priorityfilename> <algorithm> " +
				   "<regport> <niceport> <ackport> " +
				   "<sleepport>");
	    }
	}catch(IOException e){
	    System.err.println("OriginEmulator (main 1): " + e);
	    assert(false);
	}catch(OutOfMemoryError e){
	    System.err.println("" + e);
	    e.printStackTrace();
	    assert(false);
	}
    }

    //-----------------------------------------------------------------------
    // Generate a file, including all parent directories, if it doesn't exist
    //-----------------------------------------------------------------------
    private static void genFile(String objectID, int lockToken)
    {
	File file = null;
	File parent = null;
	boolean success = false;

	assert(!objectID.endsWith("/"));
	file = new File(objectID);
	parent = file.getParentFile();
	if(!parent.exists()){
	    success = parent.mkdirs();
	    assert(success);
	}
	parent = null;
	file = null;
    }

    //-----------------------------------------------------------------------
    // Fill a file with random data
    //-----------------------------------------------------------------------
    private static void fillFile(String objectID, long size, int lockToken)
	throws IOException
    {
	double rand = 0;
	byte[] buffer = null;
	long numBytesRemaining = 0;
	FileOutputStream fileOS = null;

	try{
	    numBytesRemaining = size;
	    buffer = new byte[1024];
	    fileOS = new FileOutputStream(objectID);
	    while(numBytesRemaining > 0){
		for(int i = 0; i < buffer.length; i++){
		    rand = Math.random() * 256.0;
		    buffer[i] = (byte)rand;
		}
		if(numBytesRemaining > buffer.length){
		    fileOS.write(buffer);
		    numBytesRemaining -= buffer.length;
		}else{
		    fileOS.write(buffer, 0, (int)numBytesRemaining);
		    numBytesRemaining = 0;
		}
	    }
	}finally{
	    if(fileOS != null){
		try{
		    fileOS.close();
		}catch(IOException e){}
		fileOS = null;
	    }
	    buffer = null;
	}
    }

    //-----------------------------------------------------------------------
    // Protected data
    //-----------------------------------------------------------------------
    protected SyncLMCommPoint lmComm;
    protected Origin origin;
    protected long offsetMS;
    protected PriorityTracker priorityTracker;
    protected BufferedReader stdinReader;
}
