import java.net.Socket;
import java.io.IOException;
import java.io.DataInput;
import java.io.DataOutput;
import java.io.ByteArrayOutputStream;

//---------------------------------------------------------------------------
// OriginMsg: Represents a message sent to/from the origin
//---------------------------------------------------------------------------
public class OriginMsg implements AppMessage{

    //-----------------------------------------------------------------------
    // Constructor
    //-----------------------------------------------------------------------
    public OriginMsg(long seqNum, int msgType, String objectID)
    {
	this.seqNum = seqNum;
	this.msgType = msgType;
	this.objectID = objectID;
	this.data = null;
	this.timeStamp = System.currentTimeMillis();
	assert((this.msgType >= AppMessage.READ) &&
	       (this.msgType <= AppMessage.ACK));
    }

    //-----------------------------------------------------------------------
    // Constructor
    //-----------------------------------------------------------------------
    public OriginMsg(long seqNum, int msgType, String objectID, byte[] data)
    {
	this.seqNum = seqNum;
	this.msgType = msgType;
	this.objectID = objectID;
	this.data = data;
	this.timeStamp = System.currentTimeMillis();
	assert((this.msgType >= AppMessage.READ) &&
	       (this.msgType <= AppMessage.ACK));
    }

    //-----------------------------------------------------------------------
    // Constructor
    //-----------------------------------------------------------------------
    public OriginMsg(DataInput din) throws IOException
    {
	int dataLen = 0;
	int objectIDLen = 0;
	byte[] objectIDBytes = null;

	try{
	    this.seqNum = din.readLong();
	    this.msgType = din.readInt();
	    objectIDLen = din.readInt();
	    if(objectIDLen >= 0){
		objectIDBytes = new byte[objectIDLen];
		din.readFully(objectIDBytes);
		this.objectID = new String(objectIDBytes);
	    }else{
		this.objectID = null;
	    }
	    dataLen = din.readInt();
	    if(dataLen >= 0){
		this.data = new byte[dataLen];
		din.readFully(this.data);
	    }else{
		this.data = null;
	    }
	    this.timeStamp = System.currentTimeMillis();
	    assert((this.msgType != AppMessage.DATA) || (this.data != null));
	}finally{
	    objectIDBytes = null;
	}
    }

    //-----------------------------------------------------------------------
    // Send this message over the passed-in data output stream
    //-----------------------------------------------------------------------
    public void sendSelf(DataOutput dout) throws IOException
    {
	byte[] objectIDBytes = null;

	try{
	    dout.writeLong(this.seqNum);
	    dout.writeInt(this.msgType);
	    if(this.objectID != null){
		objectIDBytes = this.objectID.getBytes();
		dout.writeInt(objectIDBytes.length);
		dout.write(objectIDBytes);
	    }else{
		dout.writeInt(-1);
	    }
	    if(this.data != null){
		dout.writeInt(this.data.length);
		dout.write(this.data);
	    }else{
		dout.writeInt(-1);
	    }
	}finally{
	    objectIDBytes = null;
	}
    }

    //-----------------------------------------------------------------------
    // Return the sequence number
    //-----------------------------------------------------------------------
    public long getSeqNum()
    {
	return(this.seqNum);
    }
 
    //-----------------------------------------------------------------------
    // Return the message type
    //-----------------------------------------------------------------------
    public int getMessageType()
    {
	return(this.msgType);
    }

    //-----------------------------------------------------------------------
    // Return the objectID
    //-----------------------------------------------------------------------
    public String getObjectID()
    {
	return(this.objectID);
    }

    //-----------------------------------------------------------------------
    // Return true if this message contains a lock (not applicable)
    //-----------------------------------------------------------------------
    public boolean containsLock()
    {
	return(false);
    }

    //-----------------------------------------------------------------------
    // Return the lock token stored in this message
    //-----------------------------------------------------------------------
    public int getLockToken()
    {
	assert(false);
	return(0);
    }

    //-----------------------------------------------------------------------
    // Return the data stored in this message
    //-----------------------------------------------------------------------
    public byte[] getData()
    {
	assert(this.data != null);
	return(this.data);
    }

    //-----------------------------------------------------------------------
    // Return the reply socket stored with this message
    //-----------------------------------------------------------------------
    public Socket getReplySocket()
    {
	assert(false);
	return(null);
    }

    //-----------------------------------------------------------------------
    // Return the time stamp for this message
    //-----------------------------------------------------------------------
    public long getTimeStamp()
    {
	assert(this.timeStamp <= System.currentTimeMillis());
	return(this.timeStamp);
    }

    //-----------------------------------------------------------------------
    // Used for testing
    //-----------------------------------------------------------------------
    public static void main(String[] argv)
    {
	OriginMsg msg = null;

	msg = new OriginMsg(0, AppMessage.READ, "RandString1");
	System.out.println("getMessageType() = " + msg.getMessageType());
	System.out.println("getObjectID() = " + msg.getObjectID());

	msg = new OriginMsg(0, AppMessage.DATA, "RandString2", new byte[1]);
	System.out.println("getMessageType() = " + msg.getMessageType());
	System.out.println("getObjectID() = " + msg.getObjectID());
	System.out.println("getData().length = " + msg.getData().length);

	msg = new OriginMsg(0, AppMessage.READ, "RandString3", new byte[2]);
	System.out.println("getMessageType() = " + msg.getMessageType());
	System.out.println("getObjectID() = " + msg.getObjectID());
	//System.out.println("getLockToken() = " + msg.getLockToken());
    }

    //-----------------------------------------------------------------------
    // Private data
    //-----------------------------------------------------------------------
    private long seqNum;
    private int msgType;
    private String objectID;
    private byte[] data;
    private long timeStamp;
}
