//---------------------------------------------------------------------------
// OriginQueueFSM: Origin message queue finite state machine
//---------------------------------------------------------------------------
// NOTES: Check the assertion that pendingInvQueue.isEmpty() => timer stopped
import java.io.IOException;

// Used for testing
import java.util.Date;

public class OriginQueueFSM{

    //-----------------------------------------------------------------------
    // Constants
    //-----------------------------------------------------------------------
    public static final int STATE_NORMAL = 1;
    public static final int STATE_ROLL_NO_DATA = 2;
    public static final int STATE_ROLL_DATA = 3;

    //-----------------------------------------------------------------------
    // Constructor
    //-----------------------------------------------------------------------
    public OriginQueueFSM(BooleanLock syncLock,
			  SyncQueue localQueue,
			  long timeout,
			  APMQSender demReqChannel,
			  APMQReceiver demInvChannel,
			  PrefetchClient prefChannel)
    {
	this.state = STATE_NORMAL;
	this.maxSeqNum = -1;
	this.localQueue = localQueue;
	this.timeout = timeout;
	this.desiredObjectID = null;
	this.desiredSeqNum = 0;
	this.demReqChannel = demReqChannel;
	this.demReqSeqNum = this.demReqChannel.getSeqNum();
	this.pendingInvQueue = new SyncQueue();
	this.prefDataPool = new PrefDataPool();
	this.oldDataPool = new OldDataPool();

	// Start threads
	this.timeoutThread = new TimeoutThread(this, syncLock);
	this.oldDataThread = new OldDataThread(this,
					       this.oldDataPool,
					       syncLock);
	this.tcpThread = new TCPThread(this, demInvChannel, syncLock);
	this.niceThread = new NiceThread(this, prefChannel, syncLock);
    }

    //-----------------------------------------------------------------------
    // Handle a timeout event
    //-----------------------------------------------------------------------
    public void handleTimeout()
    {
	OriginMsg invMsg = null;
	OriginMsg dataMsg = null;

	assert(this.state == STATE_NORMAL);
	assert(this.oldDataThread.isActive());
	// Make sure timer has turned itself off
	assert(!this.timeoutThread.isTimerSet());
	assert(!this.pendingInvQueue.isEmpty());
	this.timeoutThread.stopTimer();
	invMsg = (OriginMsg)this.pendingInvQueue.peekBlocked();
	this.pendingInvQueue.dequeue();
	dataMsg = this.prefDataPool.dequeue(invMsg.getSeqNum());
	if(dataMsg == null){
	    this.localQueue.enqueue(invMsg);
	}else{
	    this.localQueue.enqueue(dataMsg);
	}
	this.oldDataPool.removeOlderVersion(invMsg);
	this.incMaxSeqNum(invMsg.getSeqNum());
	this.acceptInvDataMatches();
	if(!this.pendingInvQueue.isEmpty()){
	    invMsg = (OriginMsg)this.pendingInvQueue.peekBlocked();
	    this.timeoutThread.startTimer(invMsg.getTimeStamp() +
					  this.timeout -
					  System.currentTimeMillis());
	}
    }

    //-----------------------------------------------------------------------
    // Handle the event where a file is prefetched
    //-----------------------------------------------------------------------
    public void handlePrefData(OriginMsg dataMsg)
    {
	switch(this.state){
	case STATE_NORMAL:
	    this.handlePrefDataNormal(dataMsg);
	    break;
	case STATE_ROLL_NO_DATA:
	    this.handlePrefDataRollNoData(dataMsg);
	    break;
	case STATE_ROLL_DATA:
	    this.handlePrefDataRollData(dataMsg);
	    break;
	default:
	    assert(false);
	    break;
	}
    }

    //-----------------------------------------------------------------------
    // Handle the event where an invalidate arrives
    //-----------------------------------------------------------------------
    public void handleInvalidate(OriginMsg invMsg)
    {
	switch(this.state){
	case STATE_NORMAL:
	    this.handleInvalidateNormal(invMsg);
	    break;
	case STATE_ROLL_NO_DATA:
	    this.handleInvalidateRollNoData(invMsg);
	    break;
	case STATE_ROLL_DATA:
	    this.handleInvalidateRollData(invMsg);
	    break;
	default:
	    assert(false);
	    break;
	}
    }

    //-----------------------------------------------------------------------
    // Handle the arrival of a demanded object
    //-----------------------------------------------------------------------
    public void handleDemandData(OriginMsg dataMsg)
    {
	OriginMsg invMsg = null;

	assert((dataMsg.getSeqNum() > this.maxSeqNum) || (this.maxSeqNum < 0));
	if(this.state == STATE_NORMAL){
	    if(!this.pendingInvQueue.isEmpty()){
		// Break into inv and data
		invMsg = new OriginMsg(dataMsg.getSeqNum(),
				       AppMessage.INV,
				       dataMsg.getObjectID());
		this.pendingInvQueue.enqueue(invMsg);
		this.prefDataPool.enqueue(dataMsg);
	    }else{
		// Apply it directly
		assert(!this.timeoutThread.isTimerSet());
		assert(this.oldDataThread.isActive());
		this.localQueue.enqueue(dataMsg);
		this.oldDataPool.removeOlderVersion(dataMsg);
		this.incMaxSeqNum(dataMsg.getSeqNum());
	    }
	}else{
	    // Roll mode
	    assert(!this.oldDataThread.isActive());
	    assert(!this.timeoutThread.isTimerSet());
	    assert(this.pendingInvQueue.isEmpty());
	    if(this.desiredObjectID.equals(dataMsg.getObjectID())){
		this.localQueue.enqueue(dataMsg);
		// Next line just to maintain the assertion that we
		// never return a message with seqNum < desiredSeqNum.
		// Otherwise it isn't strictly needed
		this.desiredSeqNum = dataMsg.getSeqNum();
		this.oldDataPool.removeOlderVersion(dataMsg);
		this.incMaxSeqNum(dataMsg.getSeqNum());
		this.oldDataThread.enable();
		this.state = STATE_NORMAL;
	    }else{
		// Break into inv and data
		invMsg = new OriginMsg(dataMsg.getSeqNum(),
				       AppMessage.INV,
				       dataMsg.getObjectID());
		this.localQueue.enqueue(invMsg);
		//this.prefDataPool.enqueue(dataMsg);
		this.oldDataPool.enqueue(dataMsg);
		this.incMaxSeqNum(dataMsg.getSeqNum());
	    }
	}
    }

    //-----------------------------------------------------------------------
    // Handle the arrival of an old object
    //-----------------------------------------------------------------------
    public void handleOldData(OriginMsg dataMsg)
    {
	assert(this.state == STATE_NORMAL);
	assert(this.oldDataThread.isActive());
	assert((dataMsg.getSeqNum() <= this.maxSeqNum) ||
	       (this.maxSeqNum < 0));
	this.localQueue.enqueue(dataMsg);
    }

    //-----------------------------------------------------------------------
    // Change to roll mode; i.e., fast-forward until we get desired data
    //-----------------------------------------------------------------------
    public void startRollMode(String objectID, long minSeqNum)
    {
	OriginMsg dataMsg = null;
	OriginMsg reqMsg = null;

	// NOTE: The system cannot be in different modes for different
	// files because even if the FS is parallelized only the thread
	// that has the global lock can cause the message system to switch
	// modes and the message system is guaranteed to be in normal mode
	// before the global lock is released.
	dataMsg = this.oldDataPool.dequeue(objectID);
	if((dataMsg == null) || (dataMsg.getSeqNum() < minSeqNum)){
	    // Either no data or data too old
	    dataMsg = this.prefDataPool.peekLowestSeqNum(objectID);
	    if(dataMsg == null){
		// Send off request message
		this.demReqSeqNum++;
		reqMsg = new OriginMsg(this.demReqSeqNum,
				       AppMessage.READ,
				       objectID);
		try{
		    this.demReqChannel.enqueue(reqMsg);
		}catch(ReplicaUnreceptiveException e){
		    // NOTE: Do something about this in the future!
		    assert(false);
		}catch(IOException e){
		    // NOTE: Do something about this in the future!
		    assert(false);
		}
		this.rollUntilObjectID(objectID, minSeqNum);
	    }else{
		assert(dataMsg.getSeqNum() > this.maxSeqNum);
		this.rollUntilSeqNum(objectID, dataMsg.getSeqNum());
	    }
	}else{
	    // Found (most likely) usable data
	    this.localQueue.enqueue(dataMsg);
	    this.state = STATE_NORMAL;
	}
    }

    //-----------------------------------------------------------------------
    // Increment the current maximum sequence number
    //-----------------------------------------------------------------------
    private void incMaxSeqNum(long seqNum)
    {
	OriginMsg dataMsg = null;

	assert((seqNum == (this.maxSeqNum + 1)) || (this.maxSeqNum < 0));
	this.maxSeqNum = seqNum;
	dataMsg = this.prefDataPool.dequeue(this.maxSeqNum);
	if(dataMsg != null){
	    this.oldDataPool.enqueue(dataMsg);
	}
    }

    //-----------------------------------------------------------------------
    // Iterate through all data/seqNum pairs and pass them to the next layer
    //-----------------------------------------------------------------------
    private void acceptInvDataMatches()
    {
	boolean done = false;
	String objectID = null;
	OriginMsg invMsg = null;
	OriginMsg dataMsg = null;

	// Make sure timer is turned off while we do this
	assert(!this.timeoutThread.isTimerSet());
	done = this.pendingInvQueue.isEmpty();
	while(!done){
	    invMsg = (OriginMsg)this.pendingInvQueue.peekBlocked();
	    dataMsg = this.prefDataPool.dequeue(invMsg.getSeqNum());
	    if(dataMsg != null){
		// NOTE: Testing only!
		objectID = dataMsg.getObjectID();
		assert(objectID.equals(invMsg.getObjectID()));
		// End
		this.pendingInvQueue.dequeue();
		this.localQueue.enqueue(dataMsg);
		this.oldDataPool.removeOlderVersion(invMsg);
		this.incMaxSeqNum(invMsg.getSeqNum());
		done = this.pendingInvQueue.isEmpty();
	    }else{
		done = true;
	    }
	}
    }

    //-----------------------------------------------------------------------
    // Handle the event where a file is prefetched in the normal state
    //-----------------------------------------------------------------------
    private void handlePrefDataNormal(OriginMsg dataMsg)
    {
	long currentTime = 0;
	OriginMsg invMsg = null;

	assert(this.state == STATE_NORMAL);
	if(dataMsg.getSeqNum() <= this.maxSeqNum){
	    this.oldDataPool.enqueue(dataMsg);
	}else{
	    this.prefDataPool.enqueue(dataMsg);
	    if(!this.pendingInvQueue.isEmpty()){
		assert(this.timeoutThread.isTimerSet());
		invMsg = (OriginMsg)this.pendingInvQueue.peekBlocked();
		if(invMsg.getSeqNum() == dataMsg.getSeqNum()){
		    this.timeoutThread.stopTimer();
		    this.acceptInvDataMatches();
		    if(!this.pendingInvQueue.isEmpty()){
			invMsg = (OriginMsg)this.pendingInvQueue.peekBlocked();
			currentTime = System.currentTimeMillis();
			this.timeoutThread.startTimer(invMsg.getTimeStamp() +
						      this.timeout -
						      currentTime);
		    }
		}
	    }else{
		assert(!this.timeoutThread.isTimerSet());
	    }
	}
    }

    //-----------------------------------------------------------------------
    // Handle the event where a file is prefetched in the roll no data state
    //-----------------------------------------------------------------------
    private void handlePrefDataRollNoData(OriginMsg dataMsg)
    {
	assert(this.state == STATE_ROLL_NO_DATA);
	assert(!this.oldDataThread.isActive());
	assert(!this.timeoutThread.isTimerSet());
	assert(this.pendingInvQueue.isEmpty());
	assert((this.desiredSeqNum <= this.maxSeqNum) || (this.maxSeqNum < 0));
	if(this.desiredObjectID.equals(dataMsg.getObjectID())){
	    if(dataMsg.getSeqNum() >= this.desiredSeqNum){
		if(dataMsg.getSeqNum() <= this.maxSeqNum){
		    this.localQueue.enqueue(dataMsg);
		    this.oldDataPool.removeOlderVersion(dataMsg);
		    this.oldDataThread.enable();
		    this.state = STATE_NORMAL;
		}else{
		    // Must wait to accept this data
		    this.desiredSeqNum = dataMsg.getSeqNum();
		    this.prefDataPool.enqueue(dataMsg);
		    this.state = STATE_ROLL_DATA;
		}
	    }else{
		// Object is too old to accept
	    }
	}else{
	    // Data isn't what we're looking for
	    if(dataMsg.getSeqNum() > this.maxSeqNum){
		this.prefDataPool.enqueue(dataMsg);
	    }else{
		this.oldDataPool.enqueue(dataMsg);
	    }
	}
    }

    //-----------------------------------------------------------------------
    // Handle the event where a file is prefetched in the roll data state
    //-----------------------------------------------------------------------
    private void handlePrefDataRollData(OriginMsg dataMsg)
    {
	assert(this.state == STATE_ROLL_DATA);
	assert(!this.oldDataThread.isActive());
	assert(!this.timeoutThread.isTimerSet());
	assert(this.pendingInvQueue.isEmpty());
	assert(this.desiredSeqNum > this.maxSeqNum);
	// Check to make sure the prefetch queue time didn't go backward
	assert((!this.desiredObjectID.equals(dataMsg.getObjectID())) ||
	       (this.desiredSeqNum < dataMsg.getSeqNum()));
	if(dataMsg.getSeqNum() > this.maxSeqNum){
	    this.prefDataPool.enqueue(dataMsg);
	}else{
	    this.oldDataPool.enqueue(dataMsg);
	}
    }

    //-----------------------------------------------------------------------
    // Handle the event where an invalidate arrives in normal mode
    //-----------------------------------------------------------------------
    private void handleInvalidateNormal(OriginMsg invMsg)
    {
	OriginMsg dataMsg = null;

	assert(this.state == STATE_NORMAL);
	if(!this.pendingInvQueue.isEmpty()){
	    assert(this.timeoutThread.isTimerSet());
	    this.pendingInvQueue.enqueue(invMsg);
	}else{
	    assert(!this.timeoutThread.isTimerSet());
	    dataMsg = this.prefDataPool.dequeue(invMsg.getSeqNum());
	    if(dataMsg != null){
		this.localQueue.enqueue(dataMsg);
		this.oldDataPool.removeOlderVersion(invMsg);
		this.incMaxSeqNum(invMsg.getSeqNum());
	    }else{
		this.pendingInvQueue.enqueue(invMsg);
		Globals.debugPrintln("OriginQueueFSM: Will delay inv for: " +
				     (invMsg.getTimeStamp() +
				      this.timeout -
				      System.currentTimeMillis()) +
				     "ms");
		this.timeoutThread.startTimer(invMsg.getTimeStamp() +
					      this.timeout - 
					      System.currentTimeMillis());
	    }
	}
    }

    //-----------------------------------------------------------------------
    // Handle the event where an invalidate arrives in roll no data mode
    //-----------------------------------------------------------------------
    private void handleInvalidateRollNoData(OriginMsg invMsg)
    {
	assert(this.state == STATE_ROLL_NO_DATA);
	assert(this.pendingInvQueue.isEmpty());
	assert(!this.timeoutThread.isTimerSet());
	assert(!this.oldDataThread.isActive());
	assert((this.desiredSeqNum <= this.maxSeqNum) || (this.maxSeqNum < 0));
	if((this.desiredObjectID.equals(invMsg.getObjectID())) &&
	   (this.desiredSeqNum < invMsg.getSeqNum())){
	    this.desiredSeqNum = invMsg.getSeqNum();
	}
	this.localQueue.enqueue(invMsg);
	this.oldDataPool.removeOlderVersion(invMsg);
	this.incMaxSeqNum(invMsg.getSeqNum());
    }

    //-----------------------------------------------------------------------
    // Handle the event where an invalidate arrives in roll data mode
    //-----------------------------------------------------------------------
    private void handleInvalidateRollData(OriginMsg invMsg)
    {
	OriginMsg dataMsg = null;

	assert(this.state == STATE_ROLL_DATA);
	assert(this.pendingInvQueue.isEmpty());
	assert(!this.timeoutThread.isTimerSet());
	assert(!this.oldDataThread.isActive());
	assert(invMsg.getSeqNum() <= this.desiredSeqNum);
	if(invMsg.getSeqNum() == this.desiredSeqNum){
	    assert(this.desiredObjectID.equals(invMsg.getObjectID()));
	    dataMsg = this.prefDataPool.dequeue(this.desiredSeqNum);
	    assert(dataMsg != null);
	    this.localQueue.enqueue(dataMsg);
	    this.oldDataPool.removeOlderVersion(dataMsg);
	    this.incMaxSeqNum(this.desiredSeqNum);
	    this.oldDataThread.enable();
	    this.state = STATE_NORMAL;
	}else{
	    /*
	      NOTE: Don't know why I had this earlier; may be useful,
	            but so far it seems incorrect.
	      if((this.desiredObjectID.equals(invMsg.getObjectID())) &&
	      (this.desiredSeqNum < invMsg.getSeqNum()){
	      this.desiredSeqNum = invMsg.getSeqNum();
	      }
	    */
	    this.localQueue.enqueue(invMsg);
	    this.oldDataPool.removeOlderVersion(invMsg);
	    this.incMaxSeqNum(invMsg.getSeqNum());
	}
    }

    //-----------------------------------------------------------------------
    // Fast-forward until we get a usable object with the given ID
    //-----------------------------------------------------------------------
    private void rollUntilObjectID(String objectID, long minSeqNum)
    {
	OriginMsg invMsg = null;

	// This assert is wrong because the app can ask for this object again
	// when it sees a more recent inv.
	//assert(this.state == STATE_NORMAL);
	this.desiredObjectID = objectID;
	this.desiredSeqNum = minSeqNum;
	this.oldDataThread.disable();
	this.timeoutThread.stopTimer();
	while(!this.pendingInvQueue.isEmpty()){
	    invMsg = (OriginMsg)this.pendingInvQueue.peekBlocked();
	    this.pendingInvQueue.dequeue();
	    this.localQueue.enqueue(invMsg);
	    this.oldDataPool.removeOlderVersion(invMsg);
	    this.incMaxSeqNum(invMsg.getSeqNum());
	    if((this.desiredObjectID.equals(invMsg.getObjectID())) &&
	       (this.desiredSeqNum < invMsg.getSeqNum())){
		// We found an inval for a higher seqNum than minSeqNum
		this.desiredSeqNum = invMsg.getSeqNum();
	    }
	}
	this.state = STATE_ROLL_NO_DATA;
    }

    //-----------------------------------------------------------------------
    // Fast-forward until we get an object with the given seqNum
    //-----------------------------------------------------------------------
    private void rollUntilSeqNum(String objectID, long seqNum)
    {
	OriginMsg invMsg = null;
	OriginMsg dataMsg = null;
	boolean foundData = false;

	// This assert is wrong because the app can ask for this object again
	// when it sees a more recent inv.
	//assert(this.state == STATE_NORMAL);
	assert((seqNum > this.maxSeqNum) || (this.maxSeqNum < 0));
	this.desiredObjectID = objectID;
	this.desiredSeqNum = seqNum;
	this.oldDataThread.disable();	
	this.timeoutThread.stopTimer();
	foundData = false;
	while((!foundData) && (!this.pendingInvQueue.isEmpty())){
	    invMsg = (OriginMsg)this.pendingInvQueue.peekBlocked();
	    this.pendingInvQueue.dequeue();
	    if(invMsg.getSeqNum() == this.desiredSeqNum){
		// Fast-forwarded enough; can now accept data
		dataMsg = this.prefDataPool.dequeue(this.desiredSeqNum);
		assert(dataMsg != null);
		this.localQueue.enqueue(dataMsg);
		foundData = true;
	    }else{
		this.localQueue.enqueue(invMsg);
	    }
	    this.oldDataPool.removeOlderVersion(invMsg);
	    this.incMaxSeqNum(invMsg.getSeqNum());
	}
	if(!foundData){
	    this.state = STATE_ROLL_DATA;
	}else{
	    if(!this.pendingInvQueue.isEmpty()){
		invMsg = (OriginMsg)this.pendingInvQueue.peekBlocked();
		this.timeoutThread.startTimer(invMsg.getTimeStamp() +
					      this.timeout -
					      System.currentTimeMillis());
	    }
	    this.oldDataThread.enable();
	    this.state = STATE_NORMAL;
	}
    }

    //-----------------------------------------------------------------------
    // Protected data
    //-----------------------------------------------------------------------
    protected SyncQueue localQueue;
    protected SyncQueue pendingInvQueue;
    protected long timeout;
    protected PrefDataPool prefDataPool;
    protected OldDataPool oldDataPool;
    protected TCPThread tcpThread;
    protected NiceThread niceThread;
    protected OldDataThread oldDataThread;
    protected TimeoutThread timeoutThread;
    protected long maxSeqNum;
    protected int state;
    protected String desiredObjectID;
    protected long desiredSeqNum;
    protected long demReqSeqNum;
    protected APMQSender demReqChannel;
}
