//---------------------------------------------------------------------------
// PMsgGroupFile: Represents a file that contains a group of messages
//---------------------------------------------------------------------------
import java.io.RandomAccessFile;
import java.io.IOException;
import java.io.File;
import java.io.EOFException;

public class PMsgGroupFile{

    //-----------------------------------------------------------------------
    // Constants
    //-----------------------------------------------------------------------
    public static final int MSG_GROUP_SIZE = 4;

    // NOTE: TESTING ONLY!
    //public static final int MSG_GROUP_SIZE = 2;

    //-----------------------------------------------------------------------
    // Constructor
    //-----------------------------------------------------------------------
    public PMsgGroupFile(File file) throws IOException
    {
	OriginMsg msg = null;
	RandomAccessFile raf = null;

	this.numEntries = 0;
	this.file = file;
	assert(this.file.exists());
	this.buildIndexTable();
	this.minSeqNum = this.getSeqNumFromFileName(this.file.getName());
	this.maxSeqNum = this.minSeqNum - 1;
	if(this.numEntries > 0){
	    this.maxSeqNum = this.extractMaxSeqNum();
	}
	assert((this.minSeqNum % PMsgGroupFile.MSG_GROUP_SIZE) == 0);

	// Testing only! BEGIN:
	if(this.numEntries > 0){
	    try{
		raf = new RandomAccessFile(this.file, "rw");
		raf.seek(0);
		msg = new OriginMsg(raf);
		assert(msg.getSeqNum() == this.minSeqNum);
	    }finally{
		if(raf != null){
		    try{
			raf.close();
		    }catch(IOException e){}
		    raf = null;
		}
	    }
	}
	// END
    }

    //-----------------------------------------------------------------------
    // Constructor
    //-----------------------------------------------------------------------
    public PMsgGroupFile(File dirFile, long minSeqNum) throws IOException
    {
	assert((minSeqNum % PMsgGroupFile.MSG_GROUP_SIZE) == 0);
	this.file = new File(dirFile.getPath() +
			     "/" +
			     this.getFileNameFromSeqNum(minSeqNum));
	assert(!this.file.exists());

	this.minSeqNum = minSeqNum;
	this.maxSeqNum = minSeqNum - 1;
	this.indexTable = new long[PMsgGroupFile.MSG_GROUP_SIZE + 1];
	this.numEntries = 0;
    }

    //-----------------------------------------------------------------------
    // Return the minimum sequence number in this message group
    //-----------------------------------------------------------------------
    public synchronized long getMinSeqNum()
    {
	return(this.minSeqNum);
    }

    //-----------------------------------------------------------------------
    // Return the maximum sequence number in this message group
    //-----------------------------------------------------------------------
    public synchronized long getMaxSeqNum()
    {
	return(this.maxSeqNum);
    }

    //-----------------------------------------------------------------------
    // Add a message to this message group
    //-----------------------------------------------------------------------
    public synchronized void addMsg(OriginMsg msg) throws IOException
    {
	RandomAccessFile raf = null;

	try{
	    assert(msg.getSeqNum() == (this.maxSeqNum + 1));
	    assert((msg.getSeqNum() % PMsgGroupFile.MSG_GROUP_SIZE) ==
		   this.numEntries);
	    raf = new RandomAccessFile(this.file, "rw");
	    raf.seek(raf.length());
	    msg.sendSelf(raf);
	    this.numEntries++;
	    this.indexTable[this.numEntries] = raf.getFilePointer();
	    this.maxSeqNum++;
	    assert(this.maxSeqNum <
		   (this.minSeqNum + PMsgGroupFile.MSG_GROUP_SIZE));
	    assert(this.getBySeqNum(msg.getSeqNum()).getSeqNum() ==
		   msg.getSeqNum());
	}finally{
	    if(raf != null){
		try{
		    raf.close();
		}catch(IOException e){}
		raf = null;
	    }
	}
    }

    //-----------------------------------------------------------------------
    // Return the message with the passed-in seqNum
    //-----------------------------------------------------------------------
    public synchronized OriginMsg getBySeqNum(long seqNum) throws IOException
    {
	int offset = 0;
	OriginMsg msg = null;
	RandomAccessFile raf = null;

	try{
	    raf = new RandomAccessFile(this.file, "rw");
	    offset = (int)(seqNum - this.minSeqNum);
	    assert(offset < this.numEntries);
	    raf.seek(this.indexTable[offset]);
	    msg = new OriginMsg(raf);
	    assert(msg.getSeqNum() == seqNum);
	}finally{
	    if(raf != null){
		try{
		    raf.close();
		}catch(IOException e){}
		raf = null;
	    }
	}
	return(msg);
    }

    //-----------------------------------------------------------------------
    // Delete this message group
    //-----------------------------------------------------------------------
    public synchronized void delete() throws IOException
    {
	boolean success = false;

	success = this.file.delete();
	assert(success);
	this.indexTable = null;
	this.numEntries = 0;
	this.minSeqNum = this.getSeqNumFromFileName(this.file.getName());
	this.maxSeqNum = this.minSeqNum - 1;
    }

    //-----------------------------------------------------------------------
    // Get seqNum from the file name
    //-----------------------------------------------------------------------
    private static long getSeqNumFromFileName(String fileName)
    {
	// "msgs".length() = 4
	return(Long.parseLong(fileName.substring(4)));
    }

    //-----------------------------------------------------------------------
    // Get the file name from seqNum
    //-----------------------------------------------------------------------
    private static String getFileNameFromSeqNum(long seqNum)
    {
	// "msgs".length() = 4
	return("msgs" + seqNum);
    }

    //-----------------------------------------------------------------------
    // Read in the maximum seqNum
    //-----------------------------------------------------------------------
    private long extractMaxSeqNum() throws IOException
    {
	OriginMsg msg = null;
	RandomAccessFile raf = null;

	try{
	    raf = new RandomAccessFile(this.file, "rw");
	    raf.seek(this.indexTable[this.numEntries - 1]);
	    msg = new OriginMsg(raf);
	}finally{
	    if(raf != null){
		try{
		    raf.close();
		}catch(IOException e){}
		raf = null;
	    }
	}
	return(msg.getSeqNum());
    }

    //-----------------------------------------------------------------------
    // Build the table of file indices for fast message retrieval
    //-----------------------------------------------------------------------
    private void buildIndexTable() throws IOException
    {
	boolean done = false;
	OriginMsg msg = null;
	RandomAccessFile raf = null;

	try{
	    raf = new RandomAccessFile(this.file, "rw");
	    raf.seek(0);
	    this.indexTable = new long[PMsgGroupFile.MSG_GROUP_SIZE + 1];
	    this.numEntries = 0;
	    this.indexTable[0] = 0;
	    while(!done){
		try{
		    msg = new OriginMsg(raf);
		    this.numEntries++;
		    this.indexTable[this.numEntries] = raf.getFilePointer();
		}catch(EOFException e){
		    done = true;
		}
	    }
	}finally{
	    if(raf != null){
		try{
		    raf.close();
		}catch(IOException e){}
		raf = null;
	    }
	}
    }

    /*
    //-----------------------------------------------------------------------
    // Used for testing only
    //-----------------------------------------------------------------------
    public static void main(String[] argv)
    {
	long s = 0;
	File file = null;
	OriginMsg msg = null;
	PMsgGroupFile f = null;

	try{
	    if(Integer.parseInt(argv[0]) == 1){
		file = new File("msgs8");
		file.createNewFile();
		f = new PMsgGroupFile(file);
		System.out.println("minSeqNum = " + f.getMinSeqNum());
		System.out.println("maxSeqNum = " + f.getMaxSeqNum());
		s = f.getMaxSeqNum();
		f.addMsg(new OriginMsg(s + 1, AppMessage.READ,"obj"));
		f.addMsg(new OriginMsg(s + 2, AppMessage.READ, "obj"));
		f.addMsg(new OriginMsg(s + 3, AppMessage.READ, "obj"));
	    }else if(Integer.parseInt(argv[0]) == 2){
		file = new File("msgs8");
		f = new PMsgGroupFile(file);
		f.delete();
	    }else if(Integer.parseInt(argv[0]) == 3){
		f = new PMsgGroupFile(new File(System.getProperty("user.dir")),
				      8);
		s = f.getMaxSeqNum();
		f.addMsg(new OriginMsg(s + 1, AppMessage.READ,"obj"));
		f.addMsg(new OriginMsg(s + 2, AppMessage.READ, "obj"));
		f.addMsg(new OriginMsg(s + 3, AppMessage.READ, "obj"));
	    }else if(Integer.parseInt(argv[0]) == 4){
		file = new File("msgs8");
		f = new PMsgGroupFile(file);
		for(long i = f.getMinSeqNum(); i <= f.getMaxSeqNum(); i++){
		    msg = f.getBySeqNum(i);
		    System.out.println("Got:");
		    System.out.println("\tseqNum = " + msg.getSeqNum());
		    System.out.println("\ttype = " + msg.getMessageType());
		    System.out.println("\tobjID = " + msg.getObjectID());
		}
	    }
	}catch(IOException e){
	    System.err.println("" + e);
	}
    }
    */

    //-----------------------------------------------------------------------
    // Used as a tool that can read message group files
    //-----------------------------------------------------------------------
    public static void main(String[] argv)
    {
	long s = 0;
	File file = null;
	OriginMsg msg = null;
	PMsgGroupFile f = null;

	try{
	    if(argv.length == 1){
		file = new File(argv[0]);
		f = new PMsgGroupFile(file);
		for(long i = f.getMinSeqNum(); i <= f.getMaxSeqNum(); i++){
		    msg = f.getBySeqNum(i);
		    System.out.println("Got:");
		    System.out.println("\tseqNum = " + msg.getSeqNum());
		    System.out.println("\ttype = " + msg.getMessageType());
		    System.out.println("\tobjID = " + msg.getObjectID());
		}
	    }else{
		System.err.println("Usage: PMsgGroupFile <filename>");
	    }
	}catch(IOException e){
	    System.err.println("PMsgGroupFile (main 1): " + e);
	}
    }

    //-----------------------------------------------------------------------
    // Protected data
    //-----------------------------------------------------------------------
    protected File file;
    protected long minSeqNum;
    protected long maxSeqNum;
    protected int numEntries;
    protected long[] indexTable;
}
