//---------------------------------------------------------------------------
// PMsgStore: Implement a persistent message store
//---------------------------------------------------------------------------
// NOTE: fileNum += ... needs to be fixed so that it is incremented to
// msgGroup.getMaxSeqNum() + 1
// CORRECTION: Doesn't need to be fixed completely because file names are
// guaranteed to be aligned with MSG_GROUP_SIZE. However, we have to allow
// fileNum to be incremented through gaps in sequence numbers.
import java.io.File;
import java.io.IOException;
import java.util.Hashtable;

public class PMsgStore{

    //-----------------------------------------------------------------------
    // Constants
    //-----------------------------------------------------------------------
    public static final String MAX_SEQ_NUM_FILENAME = "MAXSEQNUM";
    public static final int UNBOUNDED = -1;
    //public static final int MAX_OUTBOUND_QUEUE = 100;
    public static final int MAX_OUTBOUND_QUEUE = UNBOUNDED;

    //-----------------------------------------------------------------------
    // Constructor
    //-----------------------------------------------------------------------
    public PMsgStore(String dirName) throws IOException
    {
	boolean success = false;
	File[] files = null;
	PMsgGroupFile msgGroup = null;

	// NOTE: Need file names to be based on message group boundaries
	assert((PMaxSeqNumFile.SEQ_NUM_STRIDE %
		PMsgGroupFile.MSG_GROUP_SIZE) == 0);
	this.dirFile = new File(dirName);
	if(!this.dirFile.exists()){
	    success = this.dirFile.mkdir();
	    assert(success);
	}
	assert(this.dirFile.isDirectory());
	this.maxSeqNumFile = new PMaxSeqNumFile(dirName +
						"/" +
						MAX_SEQ_NUM_FILENAME);
	this.minSeqNum = -1;
	this.maxSeqNum = -1;
	this.msgTable = new Hashtable();
	files = this.dirFile.listFiles(new PMsgFileFilter());
	for(int i = 0; i < files.length; i++){
	    msgGroup = new PMsgGroupFile(files[i]);
	    if((this.minSeqNum < 0) ||
	       (this.minSeqNum > msgGroup.getMinSeqNum())){
		this.minSeqNum = msgGroup.getMinSeqNum();
	    }
	    if((this.maxSeqNum < 0) ||
	       (this.maxSeqNum < msgGroup.getMinSeqNum())){
		this.maxSeqNum = msgGroup.getMaxSeqNum();
	    }
	    assert((this.minSeqNum % PMsgGroupFile.MSG_GROUP_SIZE) == 0);
	    this.msgTable.put(new Long(msgGroup.getMinSeqNum()), msgGroup);
	}
	if(files.length == 0){
	    this.minSeqNum = this.maxSeqNumFile.getNextSafeSeqNum();
	    this.maxSeqNum = this.minSeqNum - 1;
	}else{
	    assert(this.maxSeqNumFile.getNextSafeSeqNum() > this.maxSeqNum);
	}
	this.nextSeqNum = this.minSeqNum;
    }

    //-----------------------------------------------------------------------
    // Enqueue the message
    //-----------------------------------------------------------------------
    public synchronized void enqueue(OriginMsg msg)
	throws ReplicaUnreceptiveException, IOException
    {
	long seqNum = 0L;
	long fileNum = 0L;
	PMsgGroupFile msgGroup = null;

	if(((this.maxSeqNum - this.minSeqNum + 1) > MAX_OUTBOUND_QUEUE) &&
	   (MAX_OUTBOUND_QUEUE != UNBOUNDED)){
	    throw(new ReplicaUnreceptiveException());
	}
	seqNum = msg.getSeqNum();
	fileNum = (seqNum / PMsgGroupFile.MSG_GROUP_SIZE) *
	    PMsgGroupFile.MSG_GROUP_SIZE;
	msgGroup = (PMsgGroupFile)this.msgTable.get(new Long(fileNum));
	if(msgGroup == null){
	    msgGroup = new PMsgGroupFile(this.dirFile, fileNum);
	    this.msgTable.put(new Long(fileNum), msgGroup);
	    /*
	      System.out.println("PMsgStore: this.msgTable.size() = " +
	      this.msgTable.size());
	    */
	}
	msgGroup.addMsg(msg);
	assert(msg.getSeqNum() == (this.maxSeqNum + 1));
	msg = null;
	this.maxSeqNum++;
	this.maxSeqNumFile.update(this.maxSeqNum);
	this.notifyAll();
    }

    //-----------------------------------------------------------------------
    // Iterate to the next message; wait if none present
    //-----------------------------------------------------------------------
    public synchronized OriginMsg getNextMsg() throws IOException
    {
	long fileNum = 0L;
	OriginMsg msg = null;
	PMsgGroupFile msgGroup = null;

	while(this.nextSeqNum > this.maxSeqNum){
	    try{
		this.wait();
	    }catch(InterruptedException e){
		System.err.println("" + e);
	    }
	}
	fileNum = (this.nextSeqNum / PMsgGroupFile.MSG_GROUP_SIZE) *
	    PMsgGroupFile.MSG_GROUP_SIZE;
	msgGroup = (PMsgGroupFile)this.msgTable.get(new Long(fileNum));
	assert(msgGroup != null);
	msg = msgGroup.getBySeqNum(this.nextSeqNum);
	this.nextSeqNum++;
	return(msg);
    }

    //-----------------------------------------------------------------------
    // Free messages with seqNum equal to or less than the passed-in value
    //-----------------------------------------------------------------------
    public synchronized void gcUntil(long seqNum) throws IOException
    {
	// Up to AND INCLUDING seqNum will be garbage collected
	long startFileNum = 0L;
	long endFileNum = 0L;
	Long fileNumLong = null;
	PMsgGroupFile msgGroup = null;

	try{
	    assert((seqNum <= this.maxSeqNum) && (seqNum >= this.minSeqNum));
	    startFileNum = (this.minSeqNum / PMsgGroupFile.MSG_GROUP_SIZE) *
		PMsgGroupFile.MSG_GROUP_SIZE;
	    // endFileNum has the message with sequence number (seqNum + 1)
	    endFileNum = ((seqNum + 1) / PMsgGroupFile.MSG_GROUP_SIZE) *
		PMsgGroupFile.MSG_GROUP_SIZE;
	    for(long fileNum = startFileNum;
		fileNum < endFileNum;
		fileNum += PMsgGroupFile.MSG_GROUP_SIZE){
		// Note: file += ... is not incorrect (see CORRECTION above)
		fileNumLong = new Long(fileNum);
		msgGroup = (PMsgGroupFile)this.msgTable.get(fileNumLong);
		assert(msgGroup != null);
		msgGroup.delete();
		msgGroup = (PMsgGroupFile)this.msgTable.remove(fileNumLong);
		assert(msgGroup != null);
		/*
		  System.out.println("PMsgStore: Just deleted seqnum file: " +
		  fileNum);
		*/
	    }
	    this.minSeqNum = (seqNum + 1);
	    if(this.nextSeqNum < this.minSeqNum){
		this.nextSeqNum = this.minSeqNum;
	    }
	}finally{
	    fileNumLong = null;
	    msgGroup = null;
	}
    }

    //-----------------------------------------------------------------------
    // Make the iterator restart from minSeqNum
    //-----------------------------------------------------------------------
    public synchronized void reset()
    {
	this.nextSeqNum = this.minSeqNum;
    }

    //-----------------------------------------------------------------------
    // Return the maximum known sequence number
    //-----------------------------------------------------------------------
    public synchronized long getMaxSeqNum()
    {
	return(this.maxSeqNum);
    }

    //-----------------------------------------------------------------------
    // Used for testing only
    //-----------------------------------------------------------------------
    public static void main(String[] argv)
    {
	long s = 0;
	OriginMsg msg = null;
	PMsgStore msgStore = null;

	try{
	    msgStore = new PMsgStore("a");
	    if(Integer.parseInt(argv[0]) == 1){
		msgStore.enqueue(new OriginMsg(1000, AppMessage.READ,"obj"));
		msgStore.enqueue(new OriginMsg(1001, AppMessage.READ,"obj"));
		msgStore.enqueue(new OriginMsg(1002, AppMessage.READ,"obj"));
	    }else if(Integer.parseInt(argv[0]) == 2){
		// Don't actually delete any files
		msgStore.gcUntil(1000);
	    }else if(Integer.parseInt(argv[0]) == 3){
		// Delete the first message file
		msgStore.gcUntil(1001);
	    }else{
		// Try getting messages
		msg = msgStore.getNextMsg();
		System.out.println("Got message: " + msg.getSeqNum());
		msg = msgStore.getNextMsg();
		System.out.println("Got message: " + msg.getSeqNum());
		System.out.println();
		System.out.println("reset()");
		msgStore.reset();
		msg = msgStore.getNextMsg();
		System.out.println("Got message: " + msg.getSeqNum());
		msg = msgStore.getNextMsg();
		System.out.println("Got message: " + msg.getSeqNum());
	    }
	    System.out.println("minSeqNum = " + msgStore.minSeqNum);
	    System.out.println("maxSeqNum = " + msgStore.maxSeqNum);
	    /*
	      PMsgStore(String dirName);
	      void gcUntil(int seqNum);
	      OriginMsg getNextMsg();
	      void enqueue(OriginMsg msg) throws ReplicaUnreceptiveException;
	      void reset();
	      long getSeqNum();
	     */
	}catch(IOException e){
	    System.err.println("" + e);
	}catch(ReplicaUnreceptiveException e){
	    System.err.println("" + e);
	}
    }

    //-----------------------------------------------------------------------
    // Protected data
    //-----------------------------------------------------------------------
    protected File dirFile;
    protected PMaxSeqNumFile maxSeqNumFile;
    protected long minSeqNum;
    protected long maxSeqNum;  // Temporary!
    protected long nextSeqNum; // Temporary!
    protected Hashtable msgTable;
}
