//---------------------------------------------------------------------------
// PrefetchClientThread: Thread that prefetches data
//---------------------------------------------------------------------------
import java.net.Socket;
import java.io.IOException;
import java.io.InputStream;
import java.io.DataInputStream;

// For testing
import java.net.InetAddress;

public class PrefetchClientThread implements Runnable{

    //-----------------------------------------------------------------------
    // Constructor
    //-----------------------------------------------------------------------
    public PrefetchClientThread(String originServerName,
				int portNum,
				SyncQueue recvQueue)
    {
	this.originServerName = originServerName;
	this.portNum = portNum;
	this.recvQueue = recvQueue;
	this.dataIS = null;
	this.socket = null;
	(new Thread(this)).start();
    }

    //-----------------------------------------------------------------------
    // Main thread method
    //-----------------------------------------------------------------------
    public void run()
    {
	while(true){
	    this.readPrefetchData();
	}
    }

    //-----------------------------------------------------------------------
    // Read prefetched data
    //-----------------------------------------------------------------------
    private void readPrefetchData()
    {
	OriginMsg msg = null;
	InputStream is = null;

	this.makeConnection();
	try{
	    is = socket.getInputStream();
	    this.dataIS = new DataInputStream(is);
	    while(true){
		msg = new OriginMsg(this.dataIS);
		this.recvQueue.enqueue(msg);
	    }
	}catch(IOException e){
	    this.closeStreams();
	}
    }

    //-----------------------------------------------------------------------
    // Keep attempting to set up a connection until successful
    //-----------------------------------------------------------------------
    private void makeConnection()
    {
	long timeout = 0;
	long wakeTime = 0;
	InputStream is = null;

	// Borrow a constant from another file
	timeout = MQSendThread.INIT_TIMEOUT;
	while(this.socket == null){
	    try{
		this.socket = new Socket(this.originServerName, this.portNum);
		System.out.println("Connected to " + this.originServerName +
				   ":" + this.portNum);
		is = this.socket.getInputStream();
		this.dataIS = new DataInputStream(is);
	    }catch(IOException e){
		System.err.println("" + e);
		if(this.socket != null){
		    this.closeStreams();
		}
		wakeTime = System.currentTimeMillis() + timeout;
		while(wakeTime > System.currentTimeMillis()){
		    try{
			Thread.sleep(wakeTime - System.currentTimeMillis());
		    }catch(InterruptedException e1){
			// Go back to sleeping
		    }
		}
		if(timeout < MQSendThread.MAX_TIMEOUT){
		    timeout *= MQSendThread.TIMEOUT_MULT_FACTOR;
		}
	    }
	}
    }

    //-----------------------------------------------------------------------
    // Close all open streams
    //-----------------------------------------------------------------------
    private void closeStreams()
    {
	InputStream is = null;

	assert(this.socket != null);
	if(this.dataIS != null){
	    try{
		this.dataIS.close();
	    }catch(IOException e){
		// Do nothing
	    }
	}
	try{
	    is = socket.getInputStream();
	    is.close();
	}catch(IOException e){
	    // Do nothing
	}
	try{
	    socket.close();
	}catch(IOException e){
	    // Do nothing
	}
    }

    //-----------------------------------------------------------------------
    // Used for testing
    //-----------------------------------------------------------------------
    public static void main(String [] argv)
    {
	PrefetchServer server = null;
	PriorityQueue pq = null;
	PrioritySendThread sendThread = null;
	SyncLMCommPoint lmCommPoint = null;
	SyncQueue sq = null;
	OriginMsg msg = null;

	try{
	    server = new PrefetchServer(4500);
	    pq = new PriorityQueue();
	    lmCommPoint = new SyncLMCommPoint(new LMCommPoint());
	    sendThread = new PrioritySendThread(pq, lmCommPoint);

	    server.addBinding(InetAddress.getByName("el-greco"), sendThread);
	    sq = new SyncQueue();
	    new PrefetchClientThread("el-greco", 4500, sq);
	    pq.enqueue("testfile", 10, 0.5);
	    Thread.sleep(2000);
	    msg = (OriginMsg)sq.peekBlocked();
	    sq.dequeue();
	    System.out.println("Type = " + msg.getMessageType());
	    System.out.println("ObjectID = " + msg.getObjectID());
	}catch(IOException e){
	    System.err.println("" + e);
	}catch(InterruptedException e){
	    System.err.println("" + e);
	}
    }

    //-----------------------------------------------------------------------
    // Protected data
    //-----------------------------------------------------------------------
    protected String originServerName;
    protected int portNum;
    protected SyncQueue recvQueue;
    protected DataInputStream dataIS;
    protected Socket socket;
}
