//---------------------------------------------------------------------------
// PrioritySendThread: Prefetch data by decreasing priority
//---------------------------------------------------------------------------
import java.io.File;
import java.io.IOException;
import java.io.OutputStream;
import java.io.FileInputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.ByteArrayOutputStream;
import java.net.Socket;

// For testing
import java.net.ServerSocket;
import java.io.InputStream;

public class PrioritySendThread implements Runnable{

    //-----------------------------------------------------------------------
    // Constructor
    //-----------------------------------------------------------------------
    public PrioritySendThread(PriorityQueue priorityQueue,
			      SyncLMCommPoint lmComm) throws IOException
    {
	this.baos = new ByteArrayOutputStream(2000);
	this.dataOS = new DataOutputStream(this.baos);
	this.socket = null;
	this.sockOS = null;
	this.lmComm = lmComm;
	this.priorityQueue = priorityQueue;
	(new Thread(this)).start();
    }

    //-----------------------------------------------------------------------
    // Main thread method
    //-----------------------------------------------------------------------
    public void run()
    {
	try{
	    while(true){
		this.baos.reset();
		this.prefetchNextObject();
	    }
	}catch(Throwable t){
	    t.printStackTrace();
	    assert(false);
	}
    }

    //-----------------------------------------------------------------------
    // Replace the current socket with the passed-in one
    //-----------------------------------------------------------------------
    public synchronized void setSocket(Socket socket) throws IOException
    {
	if(this.socket != null){
	    this.closeStreams();
	}
	this.socket = socket;
	if(this.socket != null){
	    this.sockOS = this.socket.getOutputStream();
	}
	this.notifyAll();
    }

    //-----------------------------------------------------------------------
    // Prefetch the next object
    //-----------------------------------------------------------------------
    private synchronized void prefetchNextObject() throws IOException
    {
	byte[] data = null;
	int lockToken = 0;
	PQEntry pqEntry = null;
	AppMessage lockMsg = null;
	OriginMsg dataMsg = null;
	String objectID = null;
	byte[] dataBytes = null;

	try{
	    // Wait for a socket connection
	    while(this.socket == null){
		this.wait();
	    }
	    pqEntry = this.priorityQueue.peekNext();
	    objectID = pqEntry.getObjectID();
	    lockMsg = this.lmComm.requestSharedLock(objectID);
	    assert(lockMsg.getMessageType() ==
		   AppMessage.LMREP_GRANTED_SHARED);
	    lockToken = lockMsg.getLockToken();
	    pqEntry = this.priorityQueue.dequeue(objectID);
	    // NOTE: Demand fetch may have dequeued the object
	    if(pqEntry != null){
		// NOTE: This pqEntry need not equal the previous one
		data = this.readDisk(objectID, lockToken);
		dataMsg = new OriginMsg(pqEntry.getSeqNum(),
					AppMessage.DATA,
					objectID,
					data);
		dataMsg.sendSelf(this.dataOS);
		this.dataOS.flush();
		dataBytes = this.baos.toByteArray();
		this.sockOS.write(dataBytes);
		this.sockOS.flush();
		Globals.stats.incNiceBW(dataBytes.length);
		Globals.stats.incNiceDataBW(dataMsg);
		data = null;
		dataMsg = null;
		dataBytes = null;
		Globals.debugPrintln("PrioritySendThread: Sent off " +
				     objectID + ", " +
				     pqEntry.getSeqNum());
	    }
	    pqEntry = null;
	}catch(IOException e){
	    System.err.println("" + e);
	    this.closeStreams();
	}catch(InterruptedException e){
	    System.err.println("" + e);
	}finally{
	    // Release the lock
	    if(lockToken != 0){
		this.lmComm.releaseLock(objectID, lockToken);
	    }
	}
    }

    //-----------------------------------------------------------------------
    // Read the object from disk
    //-----------------------------------------------------------------------
    private byte[] readDisk(String objectID, int lockToken)
    {
	byte[] data = null;
	File file = null;
	FileInputStream fis = null;
	DataInputStream dis = null;

	assert(this.isSharedLock(lockToken));
	try{
	    file = new File(objectID);
	    assert(file.exists());
	    data = new byte[(int)file.length()];

	    fis = new FileInputStream(objectID);
	    dis = new DataInputStream(fis);
	    dis.readFully(data);
	}catch(IOException e){
	    assert(false);
	}finally{
	    file = null;
	    if(dis != null){
		try{
		    dis.close();
		}catch(IOException e){}
		dis = null;
	    }
	    if(fis != null){
		try{
		    fis.close();
		}catch(IOException e){}
		fis = null;
	    }
	}
	return(data);
    }

    //-----------------------------------------------------------------------
    // Close a socket and related streams
    //-----------------------------------------------------------------------
    private void closeStreams()
    {
	if(this.sockOS != null){
	    try{
		this.sockOS.close();
	    }catch(IOException e){
		// Do nothing
	    }
	    this.sockOS = null;
	}
	if(this.socket != null){
	    try{
		this.socket.close();
	    }catch(IOException e){
		// Do nothing
	    }
	    this.socket = null;
	}
    }

    //-----------------------------------------------------------------------
    // Return true if the passed-in lock is a shared lock
    //-----------------------------------------------------------------------
    private boolean isSharedLock(int lockToken)
    {
        return((lockToken != 0) && ((lockToken % 2) == 1));
    }

    //-----------------------------------------------------------------------
    // Used for testing
    //-----------------------------------------------------------------------
    public static void main(String[] argv)
    {
        PriorityQueue pq = null;
	PrioritySendThread pst = null;

	try{
	    new PSTDummyServer(4505);
	    pq = new PriorityQueue();
	    pst = new PrioritySendThread(pq,
					 new SyncLMCommPoint(new
					     LMCommPoint()));
	    pst.setSocket(new Socket("localhost", 4505));
	    pq.enqueue("testfile", 10, 0.5);
	}catch(IOException e){
	    System.err.println("" + e);
	}
    }

    //-----------------------------------------------------------------------
    // Protected data
    //-----------------------------------------------------------------------
    protected ByteArrayOutputStream baos;
    protected DataOutputStream dataOS;
    protected Socket socket;
    protected OutputStream sockOS;
    protected SyncLMCommPoint lmComm;
    protected PriorityQueue priorityQueue;
}

//---------------------------------------------------------------------------
// PSTDummyServer: Used to test the PrioritySendThread
//---------------------------------------------------------------------------
class PSTDummyServer implements Runnable{

    //-----------------------------------------------------------------------
    // Constructor
    //-----------------------------------------------------------------------
    public PSTDummyServer(int portNum)
    {
	this.portNum = portNum;
	(new Thread(this)).start();
    }

    //-----------------------------------------------------------------------
    // Main thread method
    //-----------------------------------------------------------------------
    public void run()
    {
	Socket s = null;
	ServerSocket ss = null;
	InputStream is = null;
	DataInputStream dis = null;
	OriginMsg msg = null;
	byte[] data = null;

	try{
	    ss = new ServerSocket(this.portNum);
	    s = ss.accept();
	    is = s.getInputStream();
	    dis = new DataInputStream(is);

	    while(true){
		msg = new OriginMsg(dis);
		System.out.println("Msg type = " + msg.getMessageType());
		System.out.println("Msg objectID = " + msg.getObjectID());
		data = msg.getData();
		System.out.println("Msg data len = " + data.length);
	    }
	}catch(IOException e){
	    System.err.println("" + e);
	}
    }

    //-----------------------------------------------------------------------
    // Protected data
    //-----------------------------------------------------------------------
    protected int portNum;
}
