//---------------------------------------------------------------------------
// ReplicaFSM: Replica finite state machine
//---------------------------------------------------------------------------
import java.io.IOException;

public class ReplicaFSM{

    //-----------------------------------------------------------------------
    // Constructor
    //-----------------------------------------------------------------------
    public ReplicaFSM(String serverName,
		      int tcpPort,
		      int ackPort,
		      int nicePort,
		      long timeout)
    {
	this.fileStateTable = new FileStateTable();
	this.localFS = new LocalFS();
	this.lmCommPoint = new LMCommPoint();
	this.serverName = serverName;
	this.tcpPort = tcpPort;
	this.ackPort = ackPort;
	this.nicePort = nicePort;
	this.timeout = timeout;
    }

    //-----------------------------------------------------------------------
    // Start accepting and processing messages
    //-----------------------------------------------------------------------
    public void startReplica() throws IOException
    {
	BooleanLock syncLock = null;

	syncLock = new BooleanLock();
	this.originMsgQueue = new OriginMsgQueue(this,
						 this.serverName,
						 this.tcpPort,
						 this.ackPort,
						 this.nicePort,
						 this.timeout,
						 syncLock);
    }

    //-----------------------------------------------------------------------
    // Apply the given message to the state machine
    //-----------------------------------------------------------------------
    public synchronized void applyMsg(AppMessage msg)
    {
	String objectID = null;
	FileState fileStateRef = null;

	Globals.debugPrintln("ReplicaFSM: Got this message:");
	Globals.debugPrintln("Time = " + System.currentTimeMillis());
	Globals.debugPrintln("MsgType = " + msg.getMessageType());
	if((msg.getMessageType() == AppMessage.DATA) ||
	   (msg.getMessageType() == AppMessage.INV)){
	    Globals.debugPrintln("SeqNum = " + msg.getSeqNum());
	    Globals.debugPrintln("TimeStamp = " + msg.getTimeStamp());
	}
	Globals.debugPrintln("ObjectID = " + msg.getObjectID());
	Globals.debugPrintln("");

	try{
	    objectID = msg.getObjectID();
	    fileStateRef = this.fileStateTable.get(objectID);
	    switch(fileStateRef.getStateName()){
	    case FileState.STATE_IDLE:
		this.handleIdleState(msg, fileStateRef);
		break;
	    case FileState.STATE_GOT_DATA:
		this.handleGotDataState(msg, fileStateRef);
		break;
	    case FileState.STATE_GOT_INV:
		this.handleGotInvState(msg, fileStateRef);
		break;
	    case FileState.STATE_GOT_READ_AND_LOCK:
		this.handleGotReadAndLockState(msg, fileStateRef);
		break;
	    default:
		assert(false);
		break;
	    }
	    if(fileStateRef.getStateName() == FileState.STATE_IDLE){	
		this.fileStateTable.remove(objectID);
	    }
	}catch(IOException e){
	    System.err.println("ReplicaFSM (applyMsg 1): " + e);
	    e.printStackTrace();
	    assert(false);
	}
    }

    //-----------------------------------------------------------------------
    // Handle the message when in an idle state
    //-----------------------------------------------------------------------
    private void handleIdleState(AppMessage msg, FileState fileStateRef)
    {
	long seqNum = 0;
	LMMailBox mBox = null;

	switch(msg.getMessageType()){
	case AppMessage.DATA:
	    // assert(Do not have lock);
	    assert(!msg.containsLock());
	    fileStateRef.setSeqNum(msg.getSeqNum());
	    fileStateRef.fillBuffer(msg.getData());
	    mBox = new LMMailBox();
	    this.lmCommPoint.sendExclusiveLockRequest(msg.getObjectID(), mBox);
	    new LockBlockThread(this, mBox);
	    fileStateRef.setStateName(FileState.STATE_GOT_DATA);
	    break;
	case AppMessage.INV:
	    assert(!msg.containsLock());
	    mBox = new LMMailBox();
	    this.lmCommPoint.sendExclusiveLockRequest(msg.getObjectID(), mBox);
	    new LockBlockThread(this, mBox);
	    fileStateRef.setSeqNum(msg.getSeqNum());
	    fileStateRef.setStateName(FileState.STATE_GOT_INV);
	    break;
	case AppMessage.LMREP_GRANTED_EXCLUSIVE:
	    assert(msg.containsLock());
	    this.lmCommPoint.releaseLock(msg.getObjectID(),
					 msg.getLockToken());
	    break;
	case AppMessage.READ:
	    assert(msg.containsLock());
	    fileStateRef.setLockToken(msg.getLockToken());
	    fileStateRef.setReplySocket(msg.getReplySocket());
	    seqNum = this.localFS.getSeqNum(fileStateRef.getObjectID(),
					    msg.getLockToken());
	    /*
	      System.out.println("ReplicaFSM: Sending request for: " +
	      fileStateRef.getObjectID());
	    */
	    this.originMsgQueue.sendObjectRequest(fileStateRef.getObjectID(),
						  seqNum);
	    fileStateRef.setStateName(FileState.STATE_GOT_READ_AND_LOCK);
	    break;
	case AppMessage.LMREP_GRANTED_SHARED:  // Fall through
	default:
	    assert(false);
	    break;
	}
    }

    //-----------------------------------------------------------------------
    // Handle the message when we have a pending invalidate but no lock
    //-----------------------------------------------------------------------
    private void handleGotInvState(AppMessage msg, FileState fileStateRef)
	throws IOException
    {
	switch(msg.getMessageType()){
	case AppMessage.DATA:
	    if(msg.getSeqNum() >= fileStateRef.getSeqNum()){
		// If we saw a newer Inv than this data, we can't accept it.
		fileStateRef.fillBuffer(msg.getData());
		fileStateRef.setSeqNum(msg.getSeqNum());
		fileStateRef.setStateName(FileState.STATE_GOT_DATA);
	    }
	    break;
	case AppMessage.INV:
	    // NOTE: The following line is no longer valid because the
	    // server could crash and restart with a different number
	    //assert(msg.getSeqNum() == (fileStateRef.getSeqNum() + 1));
	    assert(msg.getSeqNum() > fileStateRef.getSeqNum());
	    fileStateRef.setSeqNum(msg.getSeqNum());
	    break;
	case AppMessage.LMREP_GRANTED_EXCLUSIVE:
	    this.localFS.invalidateDisk(msg.getObjectID(),
					msg.getLockToken(),
					fileStateRef.getSeqNum());
	    this.originMsgQueue.sendAck(fileStateRef.getSeqNum());
	    this.lmCommPoint.releaseLock(msg.getObjectID(),
					 msg.getLockToken());
	    fileStateRef.setStateName(FileState.STATE_IDLE);
	    break;
	case AppMessage.READ:
	    assert(msg.containsLock());
	    this.localFS.invalidateDisk(msg.getObjectID(),
					msg.getLockToken(),
					fileStateRef.getSeqNum());
	    this.originMsgQueue.sendAck(fileStateRef.getSeqNum());
	    fileStateRef.setLockToken(msg.getLockToken());
	    fileStateRef.setReplySocket(msg.getReplySocket());
	    /*
	      System.out.println("ReplicaFSM: Sending request for: " +
	      fileStateRef.getObjectID());
	    */
	    this.originMsgQueue.sendObjectRequest(msg.getObjectID(),
						  fileStateRef.getSeqNum());
	    fileStateRef.setStateName(FileState.STATE_GOT_READ_AND_LOCK);
	    break;
	case AppMessage.LMREP_GRANTED_SHARED: // Fall through
	default:
	    assert(false);
	    break;
	}
    }

    //-----------------------------------------------------------------------
    // Handle the message when we have pending data but no lock
    //-----------------------------------------------------------------------
    private void handleGotDataState(AppMessage msg, FileState fileStateRef)
	throws IOException
    {
	long seqNum = 0;

	switch(msg.getMessageType()){
	case AppMessage.DATA:
	    //assert(Do not have lock);
	    assert(!msg.containsLock());
	    if(msg.getSeqNum() > fileStateRef.getSeqNum()){
		// We've already seen newer data, so we can't accept this
		fileStateRef.fillBuffer(msg.getData());
		fileStateRef.setSeqNum(msg.getSeqNum());
	    }
	    break;
	case AppMessage.INV:
	    //assert(Do not have lock);
	    assert(!msg.containsLock());
	    // NOTE: Lower layer will ascertain seqNum_i > seqNum_i-1
	    // NOTE: This means we cannot tolerate origin+replica failure
	    assert(msg.getSeqNum() > fileStateRef.getSeqNum());
	    fileStateRef.setSeqNum(msg.getSeqNum());
	    fileStateRef.emptyBuffer();
	    fileStateRef.setStateName(FileState.STATE_GOT_INV);
	    break;
	case AppMessage.LMREP_GRANTED_EXCLUSIVE:
	    assert(msg.containsLock());
	    seqNum = this.localFS.getSeqNum(fileStateRef.getObjectID(),
					    msg.getLockToken());
	    if(fileStateRef.getSeqNum() >= seqNum){
		// Data is more recent that what is known on disk
		this.localFS.updateDisk(msg.getObjectID(),
					fileStateRef.getBuffer(),
					msg.getLockToken(),
					fileStateRef.getSeqNum());
	    }
	    // NOTE: The next statement could have been in the "if"
	    this.originMsgQueue.sendAck(fileStateRef.getSeqNum());
	    fileStateRef.emptyBuffer();
	    this.lmCommPoint.releaseLock(msg.getObjectID(),
					 msg.getLockToken());
	    fileStateRef.setStateName(FileState.STATE_IDLE);
	    break;
	case AppMessage.READ:
	    assert(msg.containsLock());
	    seqNum = this.localFS.getSeqNum(fileStateRef.getObjectID(),
					    msg.getLockToken());
	    if(fileStateRef.getSeqNum() >= seqNum){
		// Data is more recent than what is known on disk
		this.localFS.updateDisk(msg.getObjectID(),
					fileStateRef.getBuffer(),
					msg.getLockToken(),
					fileStateRef.getSeqNum());
		fileStateRef.emptyBuffer();
		FSCommPoint.sendAck(msg.getObjectID(),
				    msg.getLockToken(),
				    msg.getReplySocket());
		fileStateRef.setStateName(FileState.STATE_IDLE);
	    }else{
		// Data is too old; send a request for it
		seqNum = fileStateRef.getSeqNum();
		/*
		  System.out.println("ReplicaFSM: Sending request for: " +
		  fileStateRef.getObjectID());
		*/
		this.originMsgQueue.sendObjectRequest(msg.getObjectID(),
						      seqNum);
		fileStateRef.setLockToken(msg.getLockToken());
		fileStateRef.setStateName(FileState.STATE_GOT_READ_AND_LOCK);
	    }
	    // NOTE: The next statement could have been in the "if"
	    this.originMsgQueue.sendAck(fileStateRef.getSeqNum());
	    break;
	default:
	    assert(false);
	    break;
	}
    }

    //-----------------------------------------------------------------------
    // Handle the message when we have a pending read and lock but no data
    //-----------------------------------------------------------------------
    private void handleGotReadAndLockState(AppMessage msg,
					   FileState fileStateRef)
	throws IOException
    {
	long seqNum = 0;

	assert(this.isExclusiveLock(fileStateRef.getLockToken()));
	switch(msg.getMessageType()){
	case AppMessage.DATA:
	    seqNum = this.localFS.getSeqNum(msg.getObjectID(),
					    fileStateRef.getLockToken());
	    if(msg.getSeqNum() >= seqNum){
		// Data is newer than what we have seen on disk
		this.localFS.updateDisk(msg.getObjectID(),
					msg.getData(),
					fileStateRef.getLockToken(),
					msg.getSeqNum());
		FSCommPoint.sendAck(msg.getObjectID(),
				    fileStateRef.getLockToken(),
				    fileStateRef.getReplySocket());
		fileStateRef.setLockToken(0);
		fileStateRef.setReplySocket(null);
		// NOTE: Next line not strictly necessary
		fileStateRef.setSeqNum(msg.getSeqNum());
		fileStateRef.setStateName(FileState.STATE_IDLE);
	    }else{
		// Have seen a more recent INV/DATA; can't accept this
		/*
		  System.out.println("ReplicaFSM: Sending request for: " +
		  fileStateRef.getObjectID());
		*/
		this.originMsgQueue.sendObjectRequest(msg.getObjectID(),
						      seqNum);
	    }
	    // NOTE: The next statement could have been in the "if"
	    this.originMsgQueue.sendAck(msg.getSeqNum());
	    break;
	case AppMessage.INV:
	    assert(!msg.containsLock());
	    fileStateRef.setSeqNum(msg.getSeqNum());
	    this.localFS.invalidateDisk(msg.getObjectID(),
					fileStateRef.getLockToken(),
					fileStateRef.getSeqNum());
	    // NOTE: The next statement may generate unnecessary messages
	    /*
	      System.out.println("ReplicaFSM: Sending request for: " +
	      fileStateRef.getObjectID());
	    */
	    this.originMsgQueue.sendObjectRequest(msg.getObjectID(),
						  fileStateRef.getSeqNum());
	    this.originMsgQueue.sendAck(fileStateRef.getSeqNum());
	    break;
	case AppMessage.LMREP_GRANTED_EXCLUSIVE:  // Fall through
	case AppMessage.LMREP_GRANTED_SHARED:  // Fall through
	case AppMessage.READ:           // Fall through
	default:
	    assert(false);
	    break;
	}
    }

    //-----------------------------------------------------------------------
    // Return true if the passed-in lock is an exclusive lock
    //-----------------------------------------------------------------------
    private boolean isExclusiveLock(int lockToken)
    {
	return((lockToken != 0) && ((lockToken % 2) == 0));
    }

    //-----------------------------------------------------------------------
    // Protected data
    //-----------------------------------------------------------------------
    protected OriginMsgQueue originMsgQueue;
    protected FileStateTable fileStateTable;
    protected LMCommInterface lmCommPoint;
    protected LocalFS localFS;
    protected String serverName;
    protected int tcpPort;
    protected int ackPort;
    protected int nicePort;
    protected long timeout;
}
