//---------------------------------------------------------------------------
// ReplicaManager: Manage one replica; one instance per replica
//---------------------------------------------------------------------------
import java.io.File;
import java.io.IOException;
import java.io.FileInputStream;
import java.io.DataInputStream;

public class ReplicaManager{

    //-----------------------------------------------------------------------
    // Constructor
    //-----------------------------------------------------------------------
    public ReplicaManager(String replicaID,
			  int regPort,
			  int nicePort,
			  int ackPort,
			  PriorityQueue priorityQueue)
	throws IOException
    {
	this.demInvQueue = new APMQSender(replicaID, regPort, ackPort);
	this.seqNum = this.demInvQueue.getSeqNum();
	this.priorityQueue = priorityQueue;
    }

    //-----------------------------------------------------------------------
    // Enqueue an invalidate to be sent
    //-----------------------------------------------------------------------
    public synchronized void enqueueInval(String objectID,
					  int lockToken,
					  double priority)
    {
	OriginMsg invMsg = null;

	this.seqNum++;
	invMsg = new OriginMsg(this.seqNum, OriginMsg.INV, objectID);
	try{
	    this.demInvQueue.enqueue(invMsg);
	}catch(ReplicaUnreceptiveException e){
	    // NOTE: Do something about this in the future
	    System.err.println("ReplicaManager (enqueueInval 1): " + e);
	    assert(false);
	}catch(IOException e){
	    // NOTE: Do something about this in the future
	    System.err.println("ReplicaManager (enqueueInval 2): " + e);
	    assert(false);
	}finally{
	    invMsg = null;
	}
    }

    //-----------------------------------------------------------------------
    // Enqueue an update to be sent
    //-----------------------------------------------------------------------
    public synchronized void enqueueUpdate(String objectID,
					   int lockToken,
					   double priority)
    {
	OriginMsg invMsg = null;

	this.seqNum++;
	invMsg = new OriginMsg(this.seqNum, OriginMsg.INV, objectID);
	try{
	    this.demInvQueue.enqueue(invMsg);
	}catch(ReplicaUnreceptiveException e){
	    // NOTE: Do something about this in the future
	    System.err.println("ReplicaManager (enqueueUpdate 1): " + e);
	    assert(false);
	}catch(IOException e){
	    // NOTE: Do something about this in the future
	    System.err.println("ReplicaManager (enqueueUpdate 2): " + e);
	    assert(false);
	}finally{
	    invMsg = null;
	}
	this.priorityQueue.enqueue(objectID, this.seqNum, priority);
    }

    //-----------------------------------------------------------------------
    // Enqueue a demand reply to be sent
    //-----------------------------------------------------------------------
    public synchronized void enqueueDemandReply(String objectID, int lockToken)
    {
	byte[] data = null;
	OriginMsg dataMsg = null;

	this.seqNum++;
	data = this.readDisk(objectID, lockToken);
	dataMsg = new OriginMsg(this.seqNum, OriginMsg.DATA, objectID, data);
	try{
	    Globals.debugPrintln("ReplicaManager: Replying with: " +
				 dataMsg.getObjectID());
	    this.demInvQueue.enqueue(dataMsg);
	    // Indicate that we no longer need to prefetch this object
	    this.priorityQueue.dequeue(dataMsg.getObjectID());
	    dataMsg = null;
	    data = null;
	}catch(ReplicaUnreceptiveException e){
	    // NOTE: Do something about this in the future
	    System.err.println("ReplicaManager (enqueueDemandReply 1): " + e);
	    assert(false);
	}catch(IOException e){
	    // NOTE: Do something about this in the future
	    System.err.println("ReplicaManager (enqueueDemandReply 2): " + e);
	    assert(false);
	}catch(Throwable t){
	    t.printStackTrace();
	    assert(false);
	}
    }

    //-----------------------------------------------------------------------
    // Read data from disk
    //-----------------------------------------------------------------------
    private byte[] readDisk(String objectID, int lockToken)
    {
        byte[] data = null;
        File file = null;
        FileInputStream fis = null;
        DataInputStream dis = null;

        try{
	    file = new File(objectID);
	    assert(file.exists());
	    data = new byte[(int)file.length()];

	    fis = new FileInputStream(objectID);
	    dis = new DataInputStream(fis);
	    dis.readFully(data);
	}catch(IOException e){
	    System.err.println("ReplicaManager (readDisk 1): " + e);
	    assert(false);
	}finally{
	    if(dis != null){
		try{
		    dis.close();
		}catch(IOException e){}
		dis = null;
	    }
	    if(fis != null){
		try{
		    fis.close();
		}catch(IOException e){}
		fis = null;
	    }
	    file = null;
	}
	return(data);
    }

    //-----------------------------------------------------------------------
    // Protected data
    //-----------------------------------------------------------------------
    protected APMQSender demInvQueue;
    protected long seqNum;
    protected PriorityQueue priorityQueue;
}
