//---------------------------------------------------------------------------
// ShadowFileMgr: Interface to access shadow file
//---------------------------------------------------------------------------
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.BufferedReader;
import java.io.StreamTokenizer;
import java.io.InputStreamReader;

public class ShadowFileMgr{

    //-----------------------------------------------------------------------
    // Constants
    //-----------------------------------------------------------------------
    public static final String SHADOW_SUFFIX = ".shadow";
    public static final int STATUS_UNKNOWN = 0;
    public static final int STATUS_INV = 1;
    public static final int STATUS_DATA = 2;
    public static final String INV_CONSTANT = "I";
    public static final String DATA_CONSTANT = "D";

    //-----------------------------------------------------------------------
    // Get a shadow entry
    //-----------------------------------------------------------------------
    public ShadowEntry getEntry(String objectID)
    {
	String fileName = null;
	ShadowEntry entry = null;

	fileName = objectID + SHADOW_SUFFIX;
	entry = this.readFile(objectID, fileName);
	return(entry);
    }

    //-----------------------------------------------------------------------
    // Save the current entry to disk
    //-----------------------------------------------------------------------
    public void saveEntry(ShadowEntry entry) throws IOException
    {
	String fileName = null;

	fileName = entry.getObjectID() + SHADOW_SUFFIX;
	this.writeFile(fileName, entry);
    }

    //-----------------------------------------------------------------------
    // Read the current shadow file
    //-----------------------------------------------------------------------
    private static ShadowEntry readFile(String objectID, String fileName)
    {
	FileInputStream fis = null;
	long seqNum = 0;
	int status = STATUS_UNKNOWN;
	ShadowEntry entry = null;
	byte[] b = null;
	int numRead = 0;
	String entryString = null;
	String numString = null;
	int si = 0;

	try{
	    fis = new FileInputStream(fileName);
	    b = new byte[100];  // NOTE: Magic number; fix later!
	    numRead = fis.read(b);
	    entryString = new String(b, 0, numRead);
	    si = entryString.indexOf(' ');
	    if(si < 0){
		status = STATUS_UNKNOWN;
	    }else if(INV_CONSTANT.equals(entryString.substring(0, si))){
		status = STATUS_INV;
	    }else if(DATA_CONSTANT.equals(entryString.substring(0, si))){
		status = STATUS_DATA;
	    }else{
		status = STATUS_UNKNOWN;
	    }

	    try{
		numString = (entryString.substring(si + 1)).trim();
		seqNum = Long.parseLong(numString);
	    }catch(NumberFormatException e){
		status = STATUS_UNKNOWN;
		seqNum = 0;
	    }

	    entry = new ShadowEntry(objectID, seqNum, status);
	}catch(IOException e){
	    entry = new ShadowEntry(objectID, 0, STATUS_UNKNOWN);
	}finally{
	    if(fis != null){
		try{
		    fis.close();
		}catch(IOException e){}
		fis = null;
	    }
	    numString = null;
	    entryString = null;
	    b = null;
	}
	return(entry);
    }

    /*
      Less optimized version?
      //-----------------------------------------------------------------------
      // Read the current shadow file
      //-----------------------------------------------------------------------
      private static ShadowEntry readFile(String fileName)
      {
      FileInputStream fis = null;
      InputStreamReader isr = null;
      BufferedReader br = null;
      StreamTokenizer st = null;
      int tokenType = 0;
      long seqNum = 0;
      int status = STATUS_UNKNOWN;
      ShadowEntry entry = null;

      try{
      fis = new FileInputStream(fileName);
      isr = new InputStreamReader(fis);
      br = new BufferedReader(isr);
      st = new StreamTokenizer(br);
      st.parseNumbers();
      tokenType = st.nextToken();
      if(tokenType != StreamTokenizer.TT_EOF){
      assert(tokenType == StreamTokenizer.TT_WORD);
      if(st.sval.equals(INV_CONSTANT)){
      status = STATUS_INV;
      }else{
      status = STATUS_DATA;
      }

      tokenType = st.nextToken();
      assert(tokenType == StreamTokenizer.TT_NUMBER);
      seqNum = (long)st.nval;
      entry = new ShadowEntry(fileName, seqNum, status);
      }else{
      entry = null;
      }
      }catch(IOException e){
      entry = new ShadowEntry(fileName, 0, STATUS_UNKNOWN);
      }finally{
      if(br != null){
      try{
      br.close();
      }catch(IOException e){}
      br = null;
      }
      if(isr != null){
      try{
      isr.close();
      }catch(IOException e){}
      isr = null;
      }
      if(fis != null){
      try{
      fis.close();
      }catch(IOException e){}
      fis = null;
      }
      }
      return(entry);
      }
    */

    //-----------------------------------------------------------------------
    // Write to the current shadow file
    //-----------------------------------------------------------------------
    private static void writeFile(String fileName, ShadowEntry entry)
	throws IOException
    {
	byte[] outBytes = null;
	String outString = null;
	FileOutputStream shadowFOS = null;

	assert((entry.getStatus() == STATUS_INV) ||
	       (entry.getStatus() == STATUS_DATA));
	try{
	    shadowFOS = new FileOutputStream(fileName);
	    if(entry.getStatus() == STATUS_INV){
		outString = INV_CONSTANT + " " + entry.getSeqNum();
	    }else{
		outString = DATA_CONSTANT + " " + entry.getSeqNum();
	    }
	    outBytes = outString.getBytes();
	    shadowFOS.write(outBytes);
	}finally{
	    outBytes = null;
	    outString = null;
	    if(shadowFOS != null){
		try{
		    shadowFOS.close();
		}catch(IOException e){}
		shadowFOS = null;
	    }
	}
    }

    /*
      Less optmized version?
      //-----------------------------------------------------------------------
      // Write to the current shadow file
      //-----------------------------------------------------------------------
      private static void writeFile(String fileName, ShadowEntry entry)
      throws IOException
      {
      FileOutputStream shadowFOS = null;
      PrintWriter shadowWriter = null;

      assert((entry.getStatus() == STATUS_INV) ||
      (entry.getStatus() == STATUS_DATA));
      try{
      shadowFOS = new FileOutputStream(fileName);
      shadowWriter = new PrintWriter(shadowFOS, true);
      if(entry.getStatus() == STATUS_INV){
      shadowWriter.println(INV_CONSTANT + " " + entry.getSeqNum());
      }else{
      shadowWriter.println(DATA_CONSTANT + " " + entry.getSeqNum());
      }
      }finally{
      if(shadowWriter != null){
      shadowWriter.close();
      shadowWriter = null;
      }
      if(shadowFOS != null){
      try{
      shadowFOS.close();
      }catch(IOException e){}
      shadowFOS = null;
      }
      }
      }
    */

    //-----------------------------------------------------------------------
    // Used for testing
    //-----------------------------------------------------------------------
    public static void main(String[] argv)
    {
	ShadowFileMgr sfm = new ShadowFileMgr();
	ShadowEntry se = null;
	ShadowEntry se2 = null;

	se = sfm.getEntry(argv[0]);
	System.out.println("getObjectID() = " + se.getObjectID());
	System.out.println("getSeqNum() = " + se.getSeqNum());
	System.out.println("getStatus() = " + se.getStatus());

	se2 = new ShadowEntry(argv[0], 50, STATUS_INV);
	try{
	    sfm.saveEntry(se2);
	}catch(IOException e){
	    System.err.println("" + e);
	}
    }
}
