//---------------------------------------------------------------------------
// TimeoutThread: In charge of generating timeout events
//---------------------------------------------------------------------------
public class TimeoutThread extends AtomicThread{

    //-----------------------------------------------------------------------
    // Constants
    //-----------------------------------------------------------------------
    public static final int STATE_ON = 1;
    public static final int STATE_OFF = 2;
    public static final int STATE_PENDING_OFF = 3;
    public static final int STATE_PENDING_RESTART = 4;

    //-----------------------------------------------------------------------
    // Constructor
    //-----------------------------------------------------------------------
    public TimeoutThread(OriginQueueFSM fsm, BooleanLock lock)
    {
	super(lock);
	this.fsm = fsm;
	this.timeout = 0;
	this.state = STATE_OFF;
	this.startSelf();
    }

    //-----------------------------------------------------------------------
    // Start the timer
    //-----------------------------------------------------------------------
    public synchronized void startTimer(long timeout)
    {
	switch(this.state){
	case STATE_ON:
	    assert(false);
	    break;
	case STATE_OFF:
	    this.timeout = timeout;
	    this.state = STATE_ON;
	    this.notifyAll();
	    break;
	case STATE_PENDING_OFF:
	    this.timeout = timeout;
	    this.state = STATE_PENDING_RESTART;
	    break;
	case STATE_PENDING_RESTART:
	    assert(false);
	    break;
	default:
	    assert(false);
	    break;
	}
    }

    //-----------------------------------------------------------------------
    // Stop the timer
    //-----------------------------------------------------------------------
    public synchronized void stopTimer()
    {
	switch(this.state){
	case STATE_ON:
	    this.state = STATE_PENDING_OFF;
	    this.notifyAll();
	    break;
	case STATE_OFF:
	    // Do nothing
	    break;
	case STATE_PENDING_OFF:
	    // Do nothing
	    break;
	case STATE_PENDING_RESTART:
	    this.state = STATE_PENDING_OFF;
	    break;
	default:
	    assert(false);
	    break;
	}
    }

    //-----------------------------------------------------------------------
    // Return true if the timer is set
    //-----------------------------------------------------------------------
    public synchronized boolean isTimerSet()
    {
	return((this.state == STATE_PENDING_RESTART) ||
	       (this.state == STATE_ON));
    }

    //-----------------------------------------------------------------------
    // Main thread method
    //-----------------------------------------------------------------------
    public void run()
    {
	try{
	    while(true){
		this.waitUntilTimeoutEvent();
		this.enterCriticalSection();
		switch(this.state){
		case STATE_ON:
		    this.state = STATE_OFF;
		    this.fsm.handleTimeout();
		    break;
		case STATE_OFF:
		    assert(false);
		    break;
		case STATE_PENDING_OFF:
		    this.state = STATE_OFF;
		    break;
		case STATE_PENDING_RESTART:
		    this.state = STATE_ON;
		    break;
		default:
		    assert(false);
		    break;
		}
		this.exitCriticalSection();
	    }
	}catch(Throwable t){
	    t.printStackTrace();
	    assert(false);
	}
    }

    //-----------------------------------------------------------------------
    // Cause thread to block until it is ready to generate a timeout event
    //-----------------------------------------------------------------------
    protected synchronized void waitUntilTimeoutEvent()
    {
	long startTime = 0;
	long currentTime = 0;

	while((this.state != STATE_ON) &&
	      (this.state != STATE_PENDING_RESTART)){
	    try{
		this.wait();
	    }catch(InterruptedException e){
		System.err.println("" + e);
	    }
	}

	startTime = System.currentTimeMillis();
	currentTime = startTime;
	Globals.debugPrintln("TimeoutThread: Will wait " +
			     (this.timeout + startTime - currentTime) +
			     "ms");
	while((currentTime < (startTime + this.timeout)) &&
	      (this.state == STATE_ON)){
	    try{
		this.wait(this.timeout + startTime - currentTime);
		currentTime = System.currentTimeMillis();
	    }catch(InterruptedException e){
		System.err.println("" + e);
	    }
	}
    }

    //-----------------------------------------------------------------------
    // Protected data
    //-----------------------------------------------------------------------
    protected OriginQueueFSM fsm;
    protected long timeout;
    protected int state;
}
