//---------------------------------------------------------------------------
// LMCommPoint: Used to communicate with the lock manager
//---------------------------------------------------------------------------
import java.net.Socket;
import java.io.IOException;
import java.io.OutputStream;
import java.util.Hashtable;

public class LMCommPoint implements LMCommInterface{

    //-----------------------------------------------------------------------
    // Constants
    //-----------------------------------------------------------------------
    public static final int LM_COMM_PORT = 5555;

    //-----------------------------------------------------------------------
    // Constructor
    //-----------------------------------------------------------------------
    public LMCommPoint()
    {
	try{
	    this.sharedTable = new Hashtable();
	    this.exclusiveTable = new Hashtable();
	    this.socket = new Socket("localhost", LM_COMM_PORT);
	}catch(IOException e){
	    System.err.println("" + e);
	    assert(false);
	}
	new LMListener(this.socket, this);
    }

    //-----------------------------------------------------------------------
    // Send the request for a shared lock to the LM
    //-----------------------------------------------------------------------
    public synchronized void sendSharedLockRequest(String objectID,
						   LMCallBack callBack)
    {
	LMRequestPacket request = null;
	OutputStream os = null;
	SyncQueue queue = null;

	try{
	    request = new LMRequestPacket(AppMessage.LMOP_ACQUIRE_SHARED,
					  0,
					  objectID,
					  this.socket);
	    os = this.socket.getOutputStream();
	    request.sendSelf(os);
	    os.flush();

	    queue = (SyncQueue)this.sharedTable.get(objectID);
	    if(queue == null){
		queue = new SyncQueue();
		this.sharedTable.put(objectID, queue);
	    }
	    queue.enqueue(callBack);
	}catch(IOException e){
	    System.err.println("LMCommPoint (requestSharedLock 1): " + e);
	    assert(false);
	}finally{
	    // Do not close these, just allow GC to decrement ref. counts
	    os = null;
	    request = null;
	    queue = null;
	}
    }

    //-----------------------------------------------------------------------
    // Request an exclusive lock on the passed-in object
    //-----------------------------------------------------------------------
    public synchronized void sendExclusiveLockRequest(String objectID,
						      LMCallBack callBack)
    {
	LMRequestPacket request = null;
	OutputStream os = null;
	SyncQueue queue = null;

	try{
	    request = new LMRequestPacket(AppMessage.LMOP_ACQUIRE_EXCLUSIVE,
					  0,
					  objectID,
					  this.socket);
	    os = this.socket.getOutputStream();
	    request.sendSelf(os);
	    os.flush();

	    queue = (SyncQueue)this.exclusiveTable.get(objectID);
	    if(queue == null){
		queue = new SyncQueue();
		this.exclusiveTable.put(objectID, queue);
	    }
	    queue.enqueue(callBack);
	}catch(IOException e){
	    System.err.println("LMCommPoint (requestExclusveLock 1): " + e);
	    assert(false);
	}finally{
	    // Do not close these, just allow GC to decrement ref. counts
	    os = null;
	    request = null;
	    queue = null;
	}
    }

    //-----------------------------------------------------------------------
    // Release the passed-in lock
    //-----------------------------------------------------------------------
    public synchronized void releaseLock(String objectID, int lockToken)
    {
	LMRequestPacket request = null;
	OutputStream os = null;

	try{
	    request = new LMRequestPacket(AppMessage.LMOP_RELEASE,
					  lockToken,
					  objectID,
					  this.socket);
	    os = this.socket.getOutputStream();
	    request.sendSelf(os);
	    os.flush();

	}catch(IOException e){
	    System.err.println("LMCommPoint (releaseLock 1): " + e);
	    assert(false);
	}finally{
	    os = null;
	    request = null;
	}
    }

    //-----------------------------------------------------------------------
    // Accept the new lock message
    //-----------------------------------------------------------------------
    public synchronized void acceptLockMessage(LMReplyPacket packet)
    {
	int lockType = 0;
	SyncQueue queue = null;
	LMCallBack callBack = null;

	lockType = packet.getMessageType();
	assert((lockType == AppMessage.LMREP_GRANTED_SHARED) ||
	       (lockType == AppMessage.LMREP_GRANTED_EXCLUSIVE));

	if(lockType == AppMessage.LMREP_GRANTED_SHARED){
	    queue = (SyncQueue)this.sharedTable.get(packet.getObjectID());
	    assert(queue != null);
	    assert(!queue.isEmpty());
	    callBack = (LMCallBack)queue.peekBlocked();
	    queue.dequeue();
	    callBack.lockArrived(packet);
	    if(queue.isEmpty()){
		this.sharedTable.remove(packet.getObjectID());
	    }
	}else{
	    queue = (SyncQueue)this.exclusiveTable.get(packet.getObjectID());
	    assert(queue != null);
	    assert(!queue.isEmpty());
	    callBack = (LMCallBack)queue.peekBlocked();
	    queue.dequeue();
	    callBack.lockArrived(packet);
	    if(queue.isEmpty()){
		this.exclusiveTable.remove(packet.getObjectID());
	    }
	}
	queue = null;
	packet = null;
    }

    //-----------------------------------------------------------------------
    // Protected data
    //-----------------------------------------------------------------------
    protected Socket socket;
    protected Hashtable sharedTable;
    protected Hashtable exclusiveTable;
}
